# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# Sign-up at https://wxpusher.zjiecode.com/
#
# Login and acquire your App Token
#   - Open the backend of the application:
#         https://wxpusher.zjiecode.com/admin/
#   - Find the appToken menu from the left menu bar, here you can reset the
#     appToken, please note that after resetting, the old appToken will be
#     invalid immediately and the call interface will fail.
from itertools import chain
import json
import re

import requests

from ..common import NotifyFormat, NotifyType
from ..locale import gettext_lazy as _
from ..url import PrivacyMode
from ..utils.parse import parse_list, validate_regex
from .base import NotifyBase

# Topics are always numerical
IS_TOPIC = re.compile(r"^\s*(?P<topic>[1-9][0-9]{0,20})\s*$")

# users always start with UID_
IS_USER = re.compile(
    r"^\s*(?P<full>(?P<prefix>UID_)(?P<user>[^\s]+))\s*$", re.I
)


WXPUSHER_RESPONSE_CODES = {
    1000: "The request was processed successfully.",
    1001: "The token provided in the request is missing.",
    1002: "The token provided in the request is incorrect or expired.",
    1003: "The body of the message was not provided.",
    1004: (
        "The user or topic you're trying to send the message to does not exist"
    ),
    1005: "The app or topic binding process failed.",
    1006: "There was an error in sending the message.",
    1007: "The message content exceeds the allowed length.",
    1008: (
        "The API call frequency is too high and the server rejected the "
        "request."
    ),
    1009: (
        "There might be other issues that are not explicitly covered by "
        "the above codes"
    ),
    1010: "The IP address making the request is not whitelisted.",
}


class WxPusherContentType:
    """Defines the different supported content types."""

    TEXT = 1
    HTML = 2
    MARKDOWN = 3


class SubscriptionType:
    # Verify Subscription Time
    UNVERIFIED = 0
    PAID_USERS = 1
    UNSUBSCRIBED = 2


class NotifyWxPusher(NotifyBase):
    """A wrapper for WxPusher Notifications."""

    # The default descriptive name associated with the Notification
    service_name = "WxPusher"

    # The services URL
    service_url = "https://wxpusher.zjiecode.com/"

    # The default protocol
    secure_protocol = "wxpusher"

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = "https://github.com/caronc/apprise/wiki/Notify_wxpusher"

    # WxPusher notification endpoint
    notify_url = "https://wxpusher.zjiecode.com/api/send/message"

    # Define object templates
    templates = ("{schema}://{token}/{targets}",)

    # Define our template tokens
    template_tokens = dict(
        NotifyBase.template_tokens,
        **{
            "token": {
                "name": _("App Token"),
                "type": "string",
                "required": True,
                "regex": (r"^AT_[^\s]+$", "i"),
                "private": True,
            },
            "target_topic": {
                "name": _("Target Topic"),
                "type": "int",
                "map_to": "targets",
            },
            "target_user": {
                "name": _("Target User ID"),
                "type": "string",
                "regex": (r"^UID_[^\s]+$", "i"),
                "map_to": "targets",
            },
            "targets": {
                "name": _("Targets"),
                "type": "list:string",
            },
        },
    )

    # Define our template arguments
    template_args = dict(
        NotifyBase.template_args,
        **{
            "to": {
                "alias_of": "targets",
            },
            "token": {
                "alias_of": "token",
            },
        },
    )

    # Used for mapping the content type to our output since Apprise supports
    # The same formats that WxPusher does.
    __content_type_map = {
        NotifyFormat.MARKDOWN: WxPusherContentType.MARKDOWN,
        NotifyFormat.TEXT: WxPusherContentType.TEXT,
        NotifyFormat.HTML: WxPusherContentType.HTML,
    }

    def __init__(self, token, targets=None, **kwargs):
        """Initialize WxPusher Object."""
        super().__init__(**kwargs)

        # App Token (associated with WxPusher account)
        self.token = validate_regex(
            token, *self.template_tokens["token"]["regex"]
        )
        if not self.token:
            msg = f"An invalid WxPusher App Token ({token}) was specified."
            self.logger.warning(msg)
            raise TypeError(msg)

        # Used for URL generation afterwards only
        self._invalid_targets = []

        # For storing what is detected
        self._users = []
        self._topics = []

        # Parse our targets
        for target in parse_list(targets):
            # Validate targets and drop bad ones:
            result = IS_USER.match(target)
            if result:
                # store valid user
                self._users.append(result["full"])
                continue

            result = IS_TOPIC.match(target)
            if result:
                # store valid topic
                self._topics.append(int(result["topic"]))
                continue

            self.logger.warning(
                f"Dropped invalid WxPusher user/topic ({target}) specified.",
            )
            self._invalid_targets.append(target)

        return

    def send(self, body, title="", notify_type=NotifyType.INFO, **kwargs):
        """Perform WxPusher Notification."""

        if not self._users and not self._topics:
            # There were no services to notify
            self.logger.warning("There were no WxPusher targets to notify")
            return False

        # Prepare our headers
        headers = {
            "User-Agent": self.app_id,
            "Content-Type": "application/json",
            "Accept": "application/json",
        }

        # Prepare our payload
        payload = {
            "appToken": self.token,
            "content": body,
            "summary": title,
            "contentType": self.__content_type_map[self.notify_format],
            "topicIds": self._topics,
            "uids": self._users,
            # unsupported at this time
            # 'verifyPay': False,
            # 'verifyPayType': 0,
            "url": None,
        }

        # Some Debug Logging
        self.logger.debug(
            f"WxPusher POST URL: {self.notify_url} "
            f"(cert_verify={self.verify_certificate})"
        )
        self.logger.debug(f"WxPusher Payload: {payload}")

        # Always call throttle before any remote server i/o is made
        self.throttle()

        try:
            r = requests.post(
                self.notify_url,
                data=json.dumps(payload).encode("utf-8"),
                headers=headers,
                verify=self.verify_certificate,
                timeout=self.request_timeout,
            )

            try:
                content = json.loads(r.content)

            except (AttributeError, TypeError, ValueError):
                # ValueError = r.content is Unparsable
                # TypeError = r.content is None
                # AttributeError = r is None
                content = {}

            # 1000 is the expected return code for a successful query
            if (
                r.status_code == requests.codes.ok
                and content
                and content.get("code") == 1000
            ):

                # We're good!
                self.logger.info(
                    "Sent WxPusher notification to %d targets.",
                    len(self._users) + len(self._topics),
                )

            else:
                error_str = (
                    content.get("msg")
                    if content
                    else (
                        WXPUSHER_RESPONSE_CODES.get(
                            content.get("code") if content else None,
                            "An unknown error occured.",
                        )
                    )
                )

                # We had a problem
                status_str = (
                    error_str
                    if error_str
                    else NotifyWxPusher.http_response_code_lookup(
                        r.status_code
                    )
                )

                self.logger.warning(
                    "Failed to send WxPusher notification, "
                    "code={}/{}: {}".format(
                        r.status_code,
                        "unk" if not content else content.get("code"),
                        status_str,
                    )
                )

                self.logger.debug(
                    f"Response Details:\r\n{content if content else r.content}"
                )

                # Mark our failure
                return False

        except requests.RequestException as e:
            self.logger.warning(
                "A Connection error occurred sending WxPusher notification."
            )
            self.logger.debug(f"Socket Exception: {e!s}")

            return False

        return True

    @property
    def url_identifier(self):
        """Returns all of the identifiers that make this URL unique from
        another simliar one.

        Targets or end points should never be identified here.
        """
        return (self.secure_protocol, self.token)

    def url(self, privacy=False, *args, **kwargs):
        """Returns the URL built dynamically based on specified arguments."""

        # Define any URL parameters
        params = self.url_parameters(privacy=privacy, *args, **kwargs)

        return "{schema}://{token}/{targets}/?{params}".format(
            schema=self.secure_protocol,
            token=self.pprint(
                self.token, privacy, mode=PrivacyMode.Secret, safe=""
            ),
            targets="/".join(
                chain(
                    [str(t) for t in self._topics],
                    self._users,
                    [
                        NotifyWxPusher.quote(x, safe="")
                        for x in self._invalid_targets
                    ],
                )
            ),
            params=NotifyWxPusher.urlencode(params),
        )

    @staticmethod
    def parse_url(url):
        """Parses the URL and returns enough arguments that can allow us to re-
        instantiate this object."""
        results = NotifyBase.parse_url(url, verify_host=False)
        if not results:
            # We're done early as we couldn't load the results
            return results

        # Get our entries; split_path() looks after unquoting content for us
        # by default
        results["targets"] = NotifyWxPusher.split_path(results["fullpath"])

        # App Token
        if "token" in results["qsd"] and len(results["qsd"]["token"]):
            # Extract the App token from an argument
            results["token"] = NotifyWxPusher.unquote(results["qsd"]["token"])
            # Any host entry defined is actually part of the path
            # store it's element (if defined)
            if results["host"]:
                results["targets"].append(
                    NotifyWxPusher.split_path(results["host"])
                )

        else:
            # The hostname is our source number
            results["token"] = NotifyWxPusher.unquote(results["host"])

        # Support the 'to' variable so that we can support rooms this way too
        # The 'to' makes it easier to use yaml configuration
        if "to" in results["qsd"] and len(results["qsd"]["to"]):
            results["targets"] += NotifyWxPusher.parse_list(
                results["qsd"]["to"]
            )

        return results



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import json
from typing import Optional, Union

from ...apprise_attachment import AppriseAttachment
from ...asset import AppriseAsset
from ...exception import AppriseInvalidData
from ...utils.base64 import base64_urldecode

try:
    from cryptography.hazmat.primitives.asymmetric import ec

    # Cryptography Support enabled
    CRYPTOGRAPHY_SUPPORT = True

except ImportError:
    # Cryptography Support disabled
    CRYPTOGRAPHY_SUPPORT = False


class WebPushSubscription:
    """WebPush Subscription."""

    # Format:
    # {
    #     "endpoint": "https://fcm.googleapis.com/fcm/send/abc123...",
    #     "keys": {
    #         "p256dh": "BNcW4oA7zq5H9TKIrA3XfKclN2fX9P_7NR...",
    #         "auth": "k9Xzm43nBGo=",
    #     }
    # }
    def __init__(self, content: Union[str, dict, None] = None) -> None:
        """Prepares a webpush object provided with content Content can be a
        dictionary, or JSON String."""

        # Our variables
        self.__endpoint = None
        self.__p256dh = None
        self.__auth = None
        self.__auth_secret = None
        self.__public_key = None

        if content is not None and not self.load(content):
            raise AppriseInvalidData("Could not load subscription")

    def load(self, content: Union[str, dict, None] = None) -> bool:
        """Performs the loading/validation of the object."""

        # Reset our variables
        self.__endpoint = None
        self.__p256dh = None
        self.__auth = None
        self.__auth_secret = None
        self.__public_key = None

        if not CRYPTOGRAPHY_SUPPORT:
            return False

        if isinstance(content, str):
            try:
                content = json.loads(content)

            except (json.decoder.JSONDecodeError, TypeError, OSError):
                # Bad data
                return False

        if not isinstance(content, dict):
            # We could not load he result set
            return False

        # Retreive our contents for validation
        endpoint = content.get("endpoint")
        if not isinstance(endpoint, str):
            return False

        try:
            p256dh = base64_urldecode(content["keys"]["p256dh"])
            if not p256dh:
                return False

            auth_secret = base64_urldecode(content["keys"]["auth"])
            if not auth_secret:
                return False

        except KeyError:
            return False

        try:
            # Store our data
            self.__public_key = ec.EllipticCurvePublicKey.from_encoded_point(
                ec.SECP256R1(),
                p256dh,
            )

        except ValueError:
            # Invalid p256dh key (Can't load Public Key)
            return False

        self.__endpoint = endpoint
        self.__p256dh = content["keys"]["p256dh"]
        self.__auth = content["keys"]["auth"]
        self.__auth_secret = auth_secret

        return True

    def write(self, path: str, indent: int = 2) -> bool:
        """Writes content to disk based on path specified.

        Content is a JSON file, so ideally you may wish to have `.json' as it's
        extension for clarity
        """
        if not self.__public_key:
            return False

        try:
            with open(path, "w", encoding="utf-8") as f:
                json.dump(self.dict, f, indent=indent)

        except (TypeError, OSError):
            # Could not write content
            return False

        return True

    @property
    def auth(self) -> Optional[str]:
        return self.__auth if self.__public_key else None

    @property
    def endpoint(self) -> Optional[str]:
        return self.__endpoint if self.__public_key else None

    @property
    def p256dh(self) -> Optional[str]:
        return self.__p256dh if self.__public_key else None

    @property
    def auth_secret(self) -> Optional[bytes]:
        return self.__auth_secret if self.__public_key else None

    @property
    def public_key(self) -> Optional["ec.EllipticCurvePublicKey"]:
        return self.__public_key

    @property
    def dict(self) -> dict:
        return (
            {
                "endpoint": self.__endpoint,
                "keys": {
                    "p256dh": self.__p256dh,
                    "auth": self.__auth,
                },
            }
            if self.__public_key
            else {
                "endpoint": "https://fcm.googleapis.com/fcm/send/abc123...",
                "keys": {
                    "p256dh": "<place public key in base64 here>",
                    "auth": "<place auth in base64 here>",
                },
            }
        )

    def json(self, indent: int = 2) -> str:
        """Returns JSON representation of the object."""
        return json.dumps(self.dict, indent=indent)

    def __bool__(self) -> bool:
        """Handle 'if' statement."""
        return bool(self.__public_key)

    def __str__(self) -> str:
        """Returns our JSON entry as a string."""
        # Return the first 16 characters of the detected endpoint subscription
        # id
        return (
            "" if not self.__endpoint else self.__endpoint.split("/")[-1][:16]
        )


class WebPushSubscriptionManager:
    """WebPush Subscription Manager."""

    # Format:
    # {
    #     "name1": {
    #         "endpoint": "https://fcm.googleapis.com/fcm/send/abc123...",
    #         "keys": {
    #             "p256dh": "BNcW4oA7zq5H9TKIrA3XfKclN2fX9P_7NR...",
    #             "auth": "k9Xzm43nBGo=",
    #         }
    #     },
    #     "name2": {
    #         "endpoint": "https://fcm.googleapis.com/fcm/send/abc123...",
    #         "keys": {
    #             "p256dh": "BNcW4oA7zq5H9TKIrA3XfKclN2fX9P_7NR...",
    #             "auth": "k9Xzm43nBGo=",
    #         }
    #     },

    # Defines the number of failures we can accept before we abort and assume
    # the file is bad
    max_load_failure_count = 3

    def __init__(self, asset: Optional["AppriseAsset"] = None) -> None:
        """Webpush Subscription Manager."""

        # Our subscriptions
        self.__subscriptions = {}

        # Prepare our Asset Object
        self.asset = (
            asset if isinstance(asset, AppriseAsset) else AppriseAsset()
        )

    def __getitem__(self, key: str) -> WebPushSubscription:
        """Returns our indexed value if it exists."""
        return self.__subscriptions[key.lower()]

    def __setitem__(
        self, name: str, subscription: Union[WebPushSubscription, str, dict]
    ) -> None:
        """Set's our object if possible."""

        if not self.add(subscription, name=name.lower()):
            raise AppriseInvalidData("Invalid subscription provided")

    def add(
        self,
        subscription: Union[WebPushSubscription, str, dict],
        name: Optional[str] = None,
    ) -> bool:
        """Add a subscription into our manager."""

        if not isinstance(subscription, WebPushSubscription):
            try:
                # Support loading our object
                subscription = WebPushSubscription(subscription)

            except AppriseInvalidData:
                return False

        if name is None:
            name = str(subscription)

        self.__subscriptions[name.lower()] = subscription
        return True

    def __bool__(self) -> bool:
        """True is returned if at least one subscription has been loaded."""
        return bool(self.__subscriptions)

    def __len__(self) -> int:
        """Returns the number of servers loaded; this includes those found
        within loaded configuration.

        This funtion nnever actually counts the Config entry themselves (if
        they exist), only what they contain.
        """
        return len(self.__subscriptions)

    def __iadd__(
        self, subscription: Union[WebPushSubscription, str, dict]
    ) -> "WebPushSubscriptionManager":

        if not self.add(subscription):
            raise AppriseInvalidData("Invalid subscription provided")

        return self

    def __contains__(self, key: str) -> bool:
        """Checks if the key exists."""
        return key.lower() in self.__subscriptions

    def clear(self) -> None:
        """Empties our server list."""
        self.__subscriptions.clear()

    @property
    def dict(self) -> dict:
        """Returns a dictionary of all entries."""
        return (
            {k: v.dict for k, v in self.__subscriptions.items()}
            if self.__subscriptions
            else {}
        )

    def load(self, path: str, byte_limit=0) -> bool:
        """Writes content to disk based on path specified.  Content is a JSON
        file, so ideally you may wish to have `.json' as it's extension for
        clarity.

        if byte_limit is zero, then we do not limit our file size, otherwise
        set this to the bytes you want to restrict yourself by
        """

        # Reset our object
        self.clear()

        # Create our attachment object
        attach = AppriseAttachment(asset=self.asset)

        # Add our path
        attach.add(path)

        if byte_limit > 0:
            # Enforce maximum file size
            attach[0].max_file_size = byte_limit

        if not attach.sync():
            return False

        try:
            # Otherwise open our path
            with open(attach[0].path, encoding="utf-8") as f:
                content = json.load(f)

        except (json.decoder.JSONDecodeError, TypeError, OSError):
            # Could not read
            return False

        if not isinstance(content, dict):
            # Not a list of dictionaries
            return False

        # Verify if we're dealing with a single element:
        # {
        #     "endpoint": "https://fcm.googleapis.com/fcm/send/abc123...",
        #     "keys": {
        #         "p256dh": "BNcW4oA7zq5H9TKIrA3XfKclN2fX9P_7NR...",
        #         "auth": "k9Xzm43nBGo=",
        #     }
        # }
        #
        # or if we're dealing with a multiple set
        #
        # {
        #     "name1": {
        #         "endpoint": "https://fcm.googleapis.com/fcm/send/abc123...",
        #         "keys": {
        #             "p256dh": "BNcW4oA7zq5H9TKIrA3XfKclN2fX9P_7NR...",
        #             "auth": "k9Xzm43nBGo=",
        #         }
        #     },
        #     "name2": {
        #         "endpoint": "https://fcm.googleapis.com/fcm/send/abc123...",
        #         "keys": {
        #             "p256dh": "BNcW4oA7zq5H9TKIrA3XfKclN2fX9P_7NR...",
        #             "auth": "k9Xzm43nBGo=",
        #         }
        #     },

        error_count = 0
        if "endpoint" in content and "keys" in content:
            if not self.add(content):
                return False

        else:
            for name, subscription in content.items():
                if not self.add(subscription, name=name.lower()):
                    error_count += 1
                    if error_count > self.max_load_failure_count:
                        self.clear()
                        return False

        return True

    def write(self, path: str, indent: int = 2) -> bool:
        """Writes content to disk based on path specified.

        Content is a JSON file, so ideally you may wish to have `.json' as it's
        extension for clarity
        """
        try:
            with open(path, "w", encoding="utf-8") as f:
                json.dump(self.dict, f, indent=indent)

        except (TypeError, OSError):
            # Could not write content
            return False

        return True

    def json(self, indent: int = 2) -> str:
        """Returns JSON representation of the object."""
        return json.dumps(self.dict, indent=indent)



# BSD 3-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2023, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Create an account https://messagebird.com if you don't already have one
#
# Get your auth_id and auth token from the dashboard here:
#   - https://console.plivo.com/dashboard/
#

from json import dumps

import requests

from ..common import NotifyType
from ..locale import gettext_lazy as _
from ..utils.parse import (
    is_phone_no,
    parse_bool,
    parse_phone_no,
    validate_regex,
)
from .base import NotifyBase


class NotifyPlivo(NotifyBase):
    """A wrapper for Plivo Notifications."""

    # The default descriptive name associated with the Notification
    service_name = "Plivo"

    # The services URL
    service_url = "https://plivo.com"

    # The default protocol
    secure_protocol = "plivo"

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = "https://github.com/caronc/apprise/wiki/Notify_plivo"

    # Plivo uses the http protocol with JSON requests
    notify_url = "https://api.plivo.com/v1/Account/{auth_id}/Message/"

    # The maximum number of messages that can be sent in a single batch
    default_batch_size = 20

    # The maximum length of the body
    body_maxlen = 140

    # A title can not be used for SMS Messages.  Setting this to zero will
    # cause any title (if defined) to get placed into the message body.
    title_maxlen = 0

    # Define object templates
    templates = (
        "{schema}://{auth_id}@{token}/{source}",
        "{schema}://{auth_id}@{token}/{source}/{targets}",
    )

    # Define our template tokens
    template_tokens = dict(
        NotifyBase.template_tokens,
        **{
            "auth_id": {
                "name": _("Auth ID"),
                "type": "string",
                "required": True,
                "regex": (r"^[a-z0-9]{20,30}$", "i"),
            },
            "token": {
                "name": _("Auth Token"),
                "type": "string",
                "required": True,
                "regex": (r"^[a-z0-9]{30,50}$", "i"),
            },
            "source": {
                "name": _("Source Phone No"),
                "type": "string",
                "prefix": "+",
                "required": True,
                "regex": (r"^[0-9\s)(+-]+$", "i"),
            },
            "target_phone": {
                "name": _("Target Phone No"),
                "type": "string",
                "prefix": "+",
                "regex": (r"^[0-9\s)(+-]+$", "i"),
                "map_to": "targets",
            },
            "targets": {
                "name": _("Targets"),
                "type": "list:string",
            },
        },
    )

    # Define our template arguments
    template_args = dict(
        NotifyBase.template_args,
        **{
            "to": {
                "alias_of": "targets",
            },
            "from": {
                "alias_of": "source",
            },
            "token": {
                "alias_of": "token",
            },
            "id": {
                "alias_of": "auth_id",
            },
            "batch": {
                "name": _("Batch Mode"),
                "type": "bool",
                "default": False,
            },
        },
    )

    def __init__(
        self, auth_id, token, source, targets=None, batch=None, **kwargs
    ):
        """Initialize Plivo Object."""
        super().__init__(**kwargs)

        self.auth_id = validate_regex(
            auth_id, *self.template_tokens["auth_id"]["regex"]
        )
        if not self.auth_id:
            msg = (
                f"The Plivo authentication ID specified ({auth_id}) is "
                "invalid."
            )
            self.logger.warning(msg)
            raise TypeError(msg)

        self.token = validate_regex(
            token, *self.template_tokens["token"]["regex"]
        )
        if not self.token:
            msg = (
                f"The Plivo authentication token specified ({token}) is "
                "invalid."
            )
            self.logger.warning(msg)
            raise TypeError(msg)

        result = is_phone_no(source)
        if not result:
            msg = f"The Plivo source specified ({source}) is invalid."
            self.logger.warning(msg)
            raise TypeError(msg)

        # Store our source; enforce E.164 format
        self.source = f'+{result["full"]}'

        # Parse our targets
        self.targets = []

        if targets:
            for target in parse_phone_no(targets):
                # Validate targets and drop bad ones:
                result = is_phone_no(target)
                if result:
                    # store valid phone number; enforce E.164 format
                    self.targets.append(f'+{result["full"]}')
                    continue

                self.logger.warning(
                    f"Dropped invalid phone # ({target}) specified.",
                )
        else:
            # No sources specified, use our own phone no
            self.targets.append(self.source)

        # Set batch
        self.batch = (
            batch
            if batch is not None
            else self.template_args["batch"]["default"]
        )

    def send(self, body, title="", notify_type=NotifyType.INFO, **kwargs):
        """Perform Plivo Notification."""

        if not self.targets:
            # There were no services to notify
            self.logger.warning("There were no Plivo targets to notify.")
            return False

        # Initialize our has_error flag
        has_error = False

        # Prepare our headers
        headers = {
            "User-Agent": self.app_id,
            "Content-Type": "application/json",
        }

        # Prepare our authentication
        auth = (self.auth_id, self.token)

        # Prepare our payload
        payload = {
            "src": self.source,
            "dst": None,
            "text": body,
        }

        # Send in batches if identified to do so
        batch_size = 1 if not self.batch else self.default_batch_size

        for index in range(0, len(self.targets), batch_size):
            # Prepare our phone no (< delimits more then one)
            payload["recipients"] = ",".join(
                self.targets[index : index + batch_size]
            )

            # Some Debug Logging
            self.logger.debug(
                "Plivo POST URL:"
                f" {self.notify_url} (cert_verify={self.verify_certificate})"
            )
            self.logger.debug(f"Plivo Payload: {payload}")

            # Always call throttle before any remote server i/o is made
            self.throttle()
            try:
                r = requests.post(
                    self.notify_url,
                    data=dumps(payload),
                    headers=headers,
                    auth=auth,
                    verify=self.verify_certificate,
                    timeout=self.request_timeout,
                )

                if r.status_code not in (
                    requests.codes.ok,
                    requests.codes.accepted,
                ):
                    # We had a problem
                    status_str = NotifyPlivo.http_response_code_lookup(
                        r.status_code
                    )

                    self.logger.warning(
                        "Failed to send {} Plivo notification{}: "
                        "{}{}error={}.".format(
                            len(self.targets[index : index + batch_size]),
                            (
                                f" to {self.targets[index]}"
                                if batch_size == 1
                                else "(s)"
                            ),
                            status_str,
                            ", " if status_str else "",
                            r.status_code,
                        )
                    )

                    self.logger.debug(f"Response Details:\r\n{r.content}")

                    # Mark our failure
                    has_error = True
                    continue

                else:
                    self.logger.info(
                        "Send {} Plivo notification{}".format(
                            len(self.targets[index : index + batch_size]),
                            (
                                f" to {self.targets[index]}"
                                if batch_size == 1
                                else "(s)"
                            ),
                        )
                    )

            except requests.RequestException as e:
                self.logger.warning(
                    f"A Connection error occured sending Plivo:{self.targets} "
                    + "notification."
                )
                self.logger.debug(f"Socket Exception: {e!s}")

                # Mark our failure
                has_error = True
                continue

        return not has_error

    @property
    def url_identifier(self):
        """Returns all of the identifiers that make this URL unique from
        another simliar one.

        Targets or end points should never be identified here.
        """
        return (
            self.secure_protocol if self.secure else self.protocol,
            self.auth_id,
            self.token,
            self.source,
        )

    def url(self, privacy=False, *args, **kwargs):
        """Returns the URL built dynamically based on specified arguments."""

        # Define any arguments set
        params = {
            "batch": "yes" if self.batch else "no",
        }

        # Extend our parameters
        params.update(self.url_parameters(privacy=privacy, *args, **kwargs))

        return (
            "{schema}://{auth_id}@{token}/{source}/{targets}/?{params}".format(
                schema=self.secure_protocol,
                auth_id=self.pprint(self.auth_id, privacy, safe=""),
                token=self.pprint(self.token, privacy, safe=""),
                source=self.source,
                targets="/".join(
                    [NotifyPlivo.quote(x, safe="+") for x in self.targets]
                ),
                params=NotifyPlivo.urlencode(params),
            )
        )

    def __len__(self):
        """Returns the number of targets associated with this notification."""
        #
        # Factor batch into calculation
        #
        return len(self.targets) if self.targets else 1

    @staticmethod
    def parse_url(url):
        """Parses the URL and returns enough arguments that can allow us to
        substantiate this object."""

        results = NotifyBase.parse_url(url, verify_host=False)
        if not results:
            # We're done early as we couldn't load the results
            return results

        # The Auth ID is in the username field
        if "id" in results["qsd"] and len(results["qsd"]["id"]):
            results["auth_id"] = NotifyPlivo.unquote(results["qsd"]["id"])

        else:
            results["auth_id"] = NotifyPlivo.unquote(results["user"])

        # Get our entries; split_path() looks after unquoting content for us
        # by default
        results["targets"] = NotifyPlivo.split_path(results["fullpath"])
        if "token" in results["qsd"] and len(results["qsd"]["token"]):
            # Store token
            results["token"] = NotifyPlivo.unquote(results["qsd"]["token"])

            # go ahead and put the host entry in the targets list
            if results["host"]:
                results["targets"].insert(
                    0, NotifyPlivo.unquote(results["host"])
                )

        else:
            # The hostname is our authentication key
            results["token"] = NotifyPlivo.unquote(results["host"])

        if "from" in results["qsd"] and len(results["qsd"]["from"]):
            results["source"] = NotifyPlivo.unquote(results["qsd"]["from"])

        else:
            try:
                # The first path entry is the source/originator
                results["source"] = results["targets"].pop(0)

            except IndexError:
                # No source specified...
                results["source"] = None
                pass

        # Support the 'to' variable so that we can support targets this way too
        # The 'to' makes it easier to use yaml configuration
        if "to" in results["qsd"] and len(results["qsd"]["to"]):
            results["targets"] += NotifyPlivo.parse_phone_no(
                results["qsd"]["to"]
            )

        # Get Batch Mode Flag
        results["batch"] = parse_bool(
            results["qsd"].get(
                "batch", NotifyPlivo.template_args["batch"]["default"]
            )
        )

        return results



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# To use this plugin, you need to create a webhook; you can read more about
# this here:
#    https://dev.outlook.com/Connectors/\
#       GetStarted#creating-messages-through-office-365-connectors-\
#           in-microsoft-teams
#
# More details are here on API Construction:
#    https://docs.microsoft.com/en-ca/outlook/actionable-messages/\
#        message-card-reference
#
# I personally created a free account at teams.microsoft.com and then
# went to the store (bottom left hand side of slack like interface).
#
# From here you can search for 'Incoming Webhook'. Once you click on it,
# you can associate the webhook with your team. At this point, you can
# optionally also assign it a name, an avatar.  Finally you'll have to
# assign it a channel it will notify.
#
# When you've completed this, it will generate you a (webhook) URL that
# looks like:
#   https://team-name.webhook.office.com/webhookb2/ \
#       abcdefgf8-2f4b-4eca-8f61-225c83db1967@abcdefg2-5a99-4849-8efc-\
#        c9e78d28e57d/IncomingWebhook/291289f63a8abd3593e834af4d79f9fe/\
#          a2329f43-0ffb-46ab-948b-c9abdad9d643
#
# Yes... The URL is that big... But it looks like this (greatly simplified):
# https://TEAM-NAME.webhook.office.com/webhookb2/ABCD/IncomingWebhook/DEFG/HIJK
#             ^                                   ^                    ^    ^
#             |                                   |                    |    |
#  These are important <--------------------------^--------------------^----^
#

# The Legacy format didn't have the team name identified and reads 'outlook'
# While this still works, consider that Microsoft will be dropping support
# for this soon, so you may need to update your IncomingWebhook. Here is
# what a legacy URL looked like:
# https://outlook.office.com/webhook/ABCD/IncomingWebhook/DEFG/HIJK
#           ^                         ^                    ^    ^
#           |                         |                    |    |
#   legacy team reference: 'outlook'  |                    |    |
#                                     |                    |    |
#  These are important <--------------^--------------------^----^
#

# You'll notice that the first token is actually 2 separated by an @ symbol
# But lets just ignore that and assume it's one great big token instead.
#
# These 3 tokens need to be placed in the URL after the Team
#   msteams://TEAM/ABCD/DEFG/HIJK
#
import json
from json.decoder import JSONDecodeError
import re

import requests

from ..apprise_attachment import AppriseAttachment
from ..common import NotifyFormat, NotifyImageSize, NotifyType
from ..locale import gettext_lazy as _
from ..utils.parse import parse_bool, validate_regex
from ..utils.templates import TemplateType, apply_template
from .base import NotifyBase


class NotifyMSTeams(NotifyBase):
    """A wrapper for Microsoft Teams Notifications."""

    # The default descriptive name associated with the Notification
    service_name = "MSTeams"

    # The services URL
    service_url = "https://teams.micrsoft.com/"

    # The default secure protocol
    secure_protocol = "msteams"

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = "https://github.com/caronc/apprise/wiki/Notify_msteams"

    # MSTeams uses the http protocol with JSON requests
    notify_url_v1 = (
        "https://outlook.office.com/webhook/"
        "{token_a}/IncomingWebhook/{token_b}/{token_c}"
    )

    # New MSTeams webhook (as of April 11th, 2021)
    notify_url_v2 = (
        "https://{team}.webhook.office.com/webhookb2/"
        "{token_a}/IncomingWebhook/{token_b}/{token_c}"
    )

    notify_url_v3 = (
        "https://{team}.webhook.office.com/webhookb2/"
        "{token_a}/IncomingWebhook/{token_b}/{token_c}/{token_d}"
    )

    # Allows the user to specify the NotifyImageSize object
    image_size = NotifyImageSize.XY_72

    # The maximum allowable characters allowed in the body per message
    body_maxlen = 1000

    # Default Notification Format
    notify_format = NotifyFormat.MARKDOWN

    # There is no reason we should exceed 35KB when reading in a JSON file.
    # If it is more than this, then it is not accepted
    max_msteams_template_size = 35000

    # Define object templates
    templates = (
        # New required format
        "{schema}://{team}/{token_a}/{token_b}/{token_c}",
        # Deprecated
        "{schema}://{token_a}/{token_b}/{token_c}",
    )

    # Define our template tokens
    template_tokens = dict(
        NotifyBase.template_tokens,
        **{
            # The Microsoft Team Name
            "team": {
                "name": _("Team Name"),
                "type": "string",
                "required": True,
                "regex": (r"^[A-Z0-9_-]+$", "i"),
            },
            # Token required as part of the API request
            #  /AAAAAAAAA@AAAAAAAAA/........./.........
            "token_a": {
                "name": _("Token A"),
                "type": "string",
                "private": True,
                "required": True,
                "regex": (r"^[A-Z0-9-]+@[A-Z0-9-]+$", "i"),
            },
            # Token required as part of the API request
            #  /................../BBBBBBBBB/..........
            "token_b": {
                "name": _("Token B"),
                "type": "string",
                "private": True,
                "required": True,
                "regex": (r"^[a-z0-9]+$", "i"),
            },
            # Token required as part of the API request
            #  /........./........./CCCCCCCCCCCCCCCCCCCCCCCC
            "token_c": {
                "name": _("Token C"),
                "type": "string",
                "private": True,
                "required": True,
                "regex": (r"^[a-z0-9-]+$", "i"),
            },
            # Token required as part of the API request
            #  /........./........./........./DDDDDDDDDDDDDDDDD
            "token_d": {
                "name": _("Token D"),
                "type": "string",
                "private": True,
                "required": False,
                "regex": (r"^V2[a-zA-Z0-9-_]+$", "i"),
            },
        },
    )

    # Define our template arguments
    template_args = dict(
        NotifyBase.template_args,
        **{
            "image": {
                "name": _("Include Image"),
                "type": "bool",
                "default": False,
                "map_to": "include_image",
            },
            "version": {
                "name": _("Version"),
                "type": "choice:int",
                "values": (1, 2, 3),
                "default": 2,
            },
            "template": {
                "name": _("Template Path"),
                "type": "string",
                "private": True,
            },
        },
    )

    # Define our token control
    template_kwargs = {
        "tokens": {
            "name": _("Template Tokens"),
            "prefix": ":",
        },
    }

    def __init__(
        self,
        token_a,
        token_b,
        token_c,
        token_d=None,
        team=None,
        version=None,
        include_image=True,
        template=None,
        tokens=None,
        **kwargs,
    ):
        """Initialize Microsoft Teams Object.

        You can optional specify a template and identify arguments you
        wish to populate your template with when posting.  Some reserved
        template arguments that can not be over-ridden are:
           `body`, `title`, and `type`.
        """
        super().__init__(**kwargs)

        try:
            self.version = int(version)

        except TypeError:
            # None was specified... take on default
            self.version = self.template_args["version"]["default"]

        except ValueError:
            # invalid content was provided; let this get caught in the next
            # validation check for the version
            self.version = None

        if self.version not in self.template_args["version"]["values"]:
            msg = f"An invalid MSTeams Version ({version}) was specified."
            self.logger.warning(msg)
            raise TypeError(msg)

        self.team = validate_regex(team)
        if not self.team:
            NotifyBase.logger.deprecate(
                "Apprise requires you to identify your Microsoft Team name as "
                "part of the URL. e.g.: "
                "msteams://TEAM-NAME/{token_a}/{token_b}/{token_c}"
            )

            # Fallback
            self.team = "outlook"

        self.token_a = validate_regex(
            token_a, *self.template_tokens["token_a"]["regex"]
        )
        if not self.token_a:
            msg = (
                f"An invalid MSTeams (first) Token ({token_a}) was specified."
            )
            self.logger.warning(msg)
            raise TypeError(msg)

        self.token_b = validate_regex(
            token_b, *self.template_tokens["token_b"]["regex"]
        )
        if not self.token_b:
            msg = (
                f"An invalid MSTeams (second) Token ({token_b}) was specified."
            )
            self.logger.warning(msg)
            raise TypeError(msg)

        self.token_c = validate_regex(
            token_c, *self.template_tokens["token_c"]["regex"]
        )
        if not self.token_c:
            msg = (
                f"An invalid MSTeams (third) Token ({token_c}) was specified."
            )
            self.logger.warning(msg)
            raise TypeError(msg)

        self.token_d = validate_regex(
            token_d, *self.template_tokens["token_d"]["regex"]
        )

        # Place a thumbnail image inline with the message body
        self.include_image = include_image

        # Our template object is just an AppriseAttachment object
        self.template = AppriseAttachment(asset=self.asset)
        if template:
            # Add our definition to our template
            self.template.add(template)
            # Enforce maximum file size
            self.template[0].max_file_size = self.max_msteams_template_size

        # Template functionality
        self.tokens = {}
        if isinstance(tokens, dict):
            self.tokens.update(tokens)

        elif tokens:
            msg = (
                "The specified MSTeams Template Tokens "
                f"({tokens}) are not identified as a dictionary."
            )
            self.logger.warning(msg)
            raise TypeError(msg)

        self.logger.deprecate(
            "Microsoft is deprecating their MSTeams webhooks on "
            "December 31, 2025. It is advised that you switch to "
            "Microsoft Power Automate (already supported by Apprise as "
            "workflows://. For more information visit: "
            "https://github.com/caronc/apprise/wiki/Notify_workflows"
        )

    def gen_payload(
        self, body, title="", notify_type=NotifyType.INFO, **kwargs
    ):
        """This function generates our payload whether it be the generic one
        Apprise generates by default, or one provided by a specified external
        template."""

        # Acquire our to-be footer icon if configured to do so
        image_url = (
            None if not self.include_image else self.image_url(notify_type)
        )

        if not self.template:
            # By default we use a generic working payload if there was
            # no template specified
            payload = {
                "@type": "MessageCard",
                "@context": "https://schema.org/extensions",
                "summary": self.app_desc,
                "themeColor": self.color(notify_type),
                "sections": [
                    {
                        "activityImage": None,
                        "activityTitle": title,
                        "text": body,
                    },
                ],
            }

            if image_url:
                payload["sections"][0]["activityImage"] = image_url

            return payload

        # If our code reaches here, then we generate ourselves the payload
        template = self.template[0]
        if not template:
            # We could not access the attachment
            self.logger.error(
                "Could not access MSTeam template"
                f" {template.url(privacy=True)}."
            )
            return False

        # Take a copy of our token dictionary
        tokens = self.tokens.copy()

        # Apply some defaults template values
        tokens["app_body"] = body
        tokens["app_title"] = title
        tokens["app_type"] = notify_type.value
        tokens["app_id"] = self.app_id
        tokens["app_desc"] = self.app_desc
        tokens["app_color"] = self.color(notify_type)
        tokens["app_image_url"] = image_url
        tokens["app_url"] = self.app_url

        # Enforce Application mode
        tokens["app_mode"] = TemplateType.JSON

        try:
            with open(template.path) as fp:
                content = json.loads(apply_template(fp.read(), **tokens))

        except OSError:
            self.logger.error(
                f"MSTeam template {template.url(privacy=True)} could not be"
                " read."
            )
            return None

        except JSONDecodeError as e:
            self.logger.error(
                f"MSTeam template {template.url(privacy=True)} contains"
                " invalid JSON."
            )
            self.logger.debug(f"JSONDecodeError: {e}")
            return None

        # Load our JSON data (if valid)
        has_error = False
        if "@type" not in content:
            self.logger.error(
                f"MSTeam template {template.url(privacy=True)} is missing"
                " @type kwarg."
            )
            has_error = True

        if "@context" not in content:
            self.logger.error(
                f"MSTeam template {template.url(privacy=True)} is missing"
                " @context kwarg."
            )
            has_error = True

        return content if not has_error else None

    def send(self, body, title="", notify_type=NotifyType.INFO, **kwargs):
        """Perform Microsoft Teams Notification."""

        headers = {
            "User-Agent": self.app_id,
            "Content-Type": "application/json",
        }

        if self.version == 1:
            notify_url = self.notify_url_v1.format(
                token_a=self.token_a,
                token_b=self.token_b,
                token_c=self.token_c,
            )

        if self.version == 2:
            notify_url = self.notify_url_v2.format(
                team=self.team,
                token_a=self.token_a,
                token_b=self.token_b,
                token_c=self.token_c,
            )
        if self.version == 3:
            notify_url = self.notify_url_v3.format(
                team=self.team,
                token_a=self.token_a,
                token_b=self.token_b,
                token_c=self.token_c,
                token_d=self.token_d,
            )

        # Generate our payload if it's possible
        payload = self.gen_payload(
            body=body, title=title, notify_type=notify_type, **kwargs
        )
        if not payload:
            # No need to present a reason; that will come from the
            # gen_payload() function itself
            return False

        self.logger.debug(
            "MSTeams POST URL:"
            f" {notify_url} (cert_verify={self.verify_certificate!r})"
        )
        self.logger.debug(f"MSTeams Payload: {payload!s}")

        # Always call throttle before any remote server i/o is made
        self.throttle()
        try:
            r = requests.post(
                notify_url,
                data=json.dumps(payload),
                headers=headers,
                verify=self.verify_certificate,
                timeout=self.request_timeout,
            )
            if r.status_code != requests.codes.ok:
                # We had a problem
                status_str = NotifyMSTeams.http_response_code_lookup(
                    r.status_code
                )

                self.logger.warning(
                    "Failed to send MSTeams notification: "
                    "{}{}error={}.".format(
                        status_str, ", " if status_str else "", r.status_code
                    )
                )

                self.logger.debug(f"Response Details:\r\n{r.content}")

                # We failed
                return False

            else:
                self.logger.info("Sent MSTeams notification.")

        except requests.RequestException as e:
            self.logger.warning(
                "A Connection error occurred sending MSTeams notification."
            )
            self.logger.debug(f"Socket Exception: {e!s}")

            # We failed
            return False

        return True

    @property
    def url_identifier(self):
        """Returns all of the identifiers that make this URL unique from
        another simliar one.

        Targets or end points should never be identified here.
        """
        return (
            self.secure_protocol,
            self.team if self.version > 1 else None,
            self.token_a,
            self.token_b,
            self.token_c,
        )

    def url(self, privacy=False, *args, **kwargs):
        """Returns the URL built dynamically based on specified arguments."""

        # Define any URL parameters
        params = {
            "image": "yes" if self.include_image else "no",
        }

        if self.version != self.template_args["version"]["default"]:
            params["version"] = str(self.version)

        if self.template:
            params["template"] = NotifyMSTeams.quote(
                self.template[0].url(), safe=""
            )

        # Extend our parameters
        params.update(self.url_parameters(privacy=privacy, *args, **kwargs))
        # Store any template entries if specified
        params.update({f":{k}": v for k, v in self.tokens.items()})

        result = None

        if self.version == 1:
            result = (
                "{schema}://{token_a}/{token_b}/{token_c}/?{params}".format(
                    schema=self.secure_protocol,
                    token_a=self.pprint(self.token_a, privacy, safe="@"),
                    token_b=self.pprint(self.token_b, privacy, safe=""),
                    token_c=self.pprint(self.token_c, privacy, safe=""),
                    params=NotifyMSTeams.urlencode(params),
                )
            )

        if self.version == 2:
            result = (
                "{schema}://{team}/{token_a}/{token_b}/{token_c}/"
                "?{params}".format(
                    schema=self.secure_protocol,
                    team=NotifyMSTeams.quote(self.team, safe=""),
                    token_a=self.pprint(self.token_a, privacy, safe=""),
                    token_b=self.pprint(self.token_b, privacy, safe=""),
                    token_c=self.pprint(self.token_c, privacy, safe=""),
                    params=NotifyMSTeams.urlencode(params),
                )
            )

        if self.version == 3:
            result = (
                "{schema}://{team}/{token_a}/{token_b}/{token_c}/"
                "{token_d}/?{params}".format(
                    schema=self.secure_protocol,
                    team=NotifyMSTeams.quote(self.team, safe=""),
                    token_a=self.pprint(self.token_a, privacy, safe=""),
                    token_b=self.pprint(self.token_b, privacy, safe=""),
                    token_c=self.pprint(self.token_c, privacy, safe=""),
                    token_d=self.pprint(self.token_d, privacy, safe=""),
                    params=NotifyMSTeams.urlencode(params),
                )
            )
        return result

    @staticmethod
    def parse_url(url):
        """Parses the URL and returns enough arguments that can allow us to re-
        instantiate this object."""

        results = NotifyBase.parse_url(url, verify_host=False)
        if not results:
            # We're done early as we couldn't load the results
            return results

        # Get unquoted entries
        entries = NotifyMSTeams.split_path(results["fullpath"])

        # Deprecated mode (backwards compatibility)
        if results.get("user"):
            # If a user was found, it's because it's still part of the first
            # token, so we concatinate them
            results["token_a"] = "{}@{}".format(
                NotifyMSTeams.unquote(results["user"]),
                NotifyMSTeams.unquote(results["host"]),
            )

        else:
            # Get the Team from the hostname
            results["team"] = NotifyMSTeams.unquote(results["host"])

            # Get the token from the path
            results["token_a"] = (
                None if not entries else NotifyMSTeams.unquote(entries.pop(0))
            )

        results["token_b"] = (
            None if not entries else NotifyMSTeams.unquote(entries.pop(0))
        )
        results["token_c"] = (
            None if not entries else NotifyMSTeams.unquote(entries.pop(0))
        )
        results["token_d"] = (
            None if not entries else NotifyMSTeams.unquote(entries.pop(0))
        )

        # Get Image
        results["include_image"] = parse_bool(
            results["qsd"].get("image", True)
        )

        # Get Team name if defined
        if "team" in results["qsd"] and results["qsd"]["team"]:
            results["team"] = NotifyMSTeams.unquote(results["qsd"]["team"])

        # Template Handling
        if "template" in results["qsd"] and results["qsd"]["template"]:
            results["template"] = NotifyMSTeams.unquote(
                results["qsd"]["template"]
            )

        # Override version if defined
        if "version" in results["qsd"] and results["qsd"]["version"]:
            results["version"] = NotifyMSTeams.unquote(
                results["qsd"]["version"]
            )

        else:
            version = 1
            if results.get("team"):
                version = 2
            if results.get("token_d"):
                version = 3
            # Set our version if not otherwise set
            results["version"] = version

        # Store our tokens
        results["tokens"] = results["qsd:"]

        return results

    @staticmethod
    def parse_native_url(url):
        """
        Legacy Support:
            https://outlook.office.com/webhook/ABCD/IncomingWebhook/DEFG/HIJK

        New Hook Support:
            https://team-name.office.com/webhook/ABCD/IncomingWebhook/DEFG/HIJK

        Newer Hook Support:
            https://team-name.office.com/webhook/ABCD/IncomingWebhook/DEFG/HIJK/V2LMNOP
        """

        # We don't need to do incredibly details token matching as the purpose
        # of this is just to detect that were dealing with an msteams url
        # token parsing will occur once we initialize the function
        result = re.match(
            r"^https?://(?P<team>[^.]+)(?P<v2a>\.webhook)?\.office\.com/"
            r"webhook(?P<v2b>b2)?/"
            r"(?P<token_a>[A-Z0-9-]+@[A-Z0-9-]+)/"
            r"IncomingWebhook/"
            r"(?P<token_b>[A-Z0-9]+)/"
            r"(?P<token_c>[A-Z0-9-]+)/"
            r"(?P<token_d>V2[A-Z0-9-_]+)/?"
            r"(?P<params>\?.+)?$",
            url,
            re.I,
        )

        if result:
            # Version 3 URL
            return NotifyMSTeams.parse_url(
                "{schema}://{team}/{token_a}/{token_b}/{token_c}/{token_d}"
                "/{params}".format(
                    schema=NotifyMSTeams.secure_protocol,
                    team=result.group("team"),
                    token_a=result.group("token_a"),
                    token_b=result.group("token_b"),
                    token_c=result.group("token_c"),
                    token_d=result.group("token_d"),
                    params=(
                        ""
                        if not result.group("params")
                        else result.group("params")
                    ),
                )
            )

        result = re.match(
            r"^https?://(?P<team>[^.]+)(?P<v2a>\.webhook)?\.office\.com/"
            r"webhook(?P<v2b>b2)?/"
            r"(?P<token_a>[A-Z0-9-]+@[A-Z0-9-]+)/"
            r"IncomingWebhook/"
            r"(?P<token_b>[A-Z0-9]+)/"
            r"(?P<token_c>[A-Z0-9-]+)/?"
            r"(?P<params>\?.+)?$",
            url,
            re.I,
        )

        if result:
            if result.group("v2a"):
                # Version 2 URL
                return NotifyMSTeams.parse_url(
                    "{schema}://{team}/{token_a}/{token_b}/{token_c}"
                    "/{params}".format(
                        schema=NotifyMSTeams.secure_protocol,
                        team=result.group("team"),
                        token_a=result.group("token_a"),
                        token_b=result.group("token_b"),
                        token_c=result.group("token_c"),
                        params=(
                            ""
                            if not result.group("params")
                            else result.group("params")
                        ),
                    )
                )
            else:
                # Version 1 URLs
                # team is also set to 'outlook' in this case
                return NotifyMSTeams.parse_url(
                    "{schema}://{token_a}/{token_b}/{token_c}/{params}".format(
                        schema=NotifyMSTeams.secure_protocol,
                        token_a=result.group("token_a"),
                        token_b=result.group("token_b"),
                        token_c=result.group("token_c"),
                        params=(
                            ""
                            if not result.group("params")
                            else result.group("params")
                        ),
                    )
                )
        return None



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import os
import re

from ..common import ConfigFormat, ContentIncludeMode
from ..locale import gettext_lazy as _
from ..utils.disk import path_decode
from .base import ConfigBase


class ConfigFile(ConfigBase):
    """A wrapper for File based configuration sources."""

    # The default descriptive name associated with the service
    service_name = _("Local File")

    # The default protocol
    protocol = "file"

    # Configuration file inclusion can only be of the same type
    allow_cross_includes = ContentIncludeMode.STRICT

    def __init__(self, path, **kwargs):
        """Initialize File Object.

        headers can be a dictionary of key/value pairs that you want to
        additionally include as part of the server headers to post with
        """
        super().__init__(**kwargs)

        # Store our file path as it was set
        self.path = path_decode(path)

        # Track the file as it was saved
        self.__original_path = os.path.normpath(path)

        # Update the config path to be relative to our file we just loaded
        self.config_path = os.path.dirname(self.path)

        return

    def url(self, privacy=False, *args, **kwargs):
        """Returns the URL built dynamically based on specified arguments."""

        # Prepare our cache value
        if isinstance(self.cache, bool) or not self.cache:
            cache = "yes" if self.cache else "no"

        else:
            cache = int(self.cache)

        # Define any URL parameters
        params = {
            "encoding": self.encoding,
            "cache": cache,
        }

        if self.config_format:
            # A format was enforced; make sure it's passed back with the url
            params["format"] = self.config_format

        return "file://{path}{params}".format(
            path=self.quote(self.__original_path),
            params=f"?{self.urlencode(params)}" if params else "",
        )

    def read(self, **kwargs):
        """Perform retrieval of the configuration based on the specified
        request."""

        response = None

        try:
            if (
                self.max_buffer_size > 0
                and os.path.getsize(self.path) > self.max_buffer_size
            ):

                # Content exceeds maximum buffer size
                self.logger.error(
                    "File size exceeds maximum allowable buffer length"
                    f" ({int(self.max_buffer_size / 1024)}KB)."
                )
                return None

        except OSError:
            # getsize() can throw this acception if the file is missing
            # and or simply isn't accessible
            self.logger.error(f"File is not accessible: {self.path}")
            return None

        # Always call throttle before any server i/o is made
        self.throttle()

        try:
            with open(self.path, encoding=self.encoding) as f:
                # Store our content for parsing
                response = f.read()

        except (ValueError, UnicodeDecodeError):
            # A result of our strict encoding check; if we receive this
            # then the file we're opening is not something we can
            # understand the encoding of..

            self.logger.error(
                f"File not using expected encoding ({self.encoding}) :"
                f" {self.path}"
            )
            return None

        except OSError:
            # IOError is present for backwards compatibility with Python
            # versions older then 3.3.  >= 3.3 throw OSError now.

            # Could not open and/or read the file; this is not a problem since
            # we scan a lot of default paths.
            self.logger.error(f"File can not be opened for read: {self.path}")
            return None

        # Detect config format based on file extension if it isn't already
        # enforced
        if (
            self.config_format is None
            and re.match(r"^.*\.ya?ml\s*$", self.path, re.I) is not None
        ):

            # YAML Filename Detected
            self.default_config_format = ConfigFormat.YAML

        # Return our response object
        return response

    @staticmethod
    def parse_url(url):
        """Parses the URL so that we can handle all different file paths and
        return it as our path object."""

        results = ConfigBase.parse_url(url, verify_host=False)
        if not results:
            # We're done early; it's not a good URL
            return results

        match = re.match(r"[a-z0-9]+://(?P<path>[^?]+)(\?.*)?", url, re.I)
        if not match:
            return None

        results["path"] = ConfigFile.unquote(match.group("path"))
        return results



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from ..locale import gettext_lazy as _
from .base import ConfigBase


class ConfigMemory(ConfigBase):
    """For information that was loaded from memory and does not persist
    anywhere."""

    # The default descriptive name associated with the service
    service_name = _("Memory")

    # The default protocol
    protocol = "memory"

    def __init__(self, content, **kwargs):
        """Initialize Memory Object.

        Memory objects just store the raw configuration in memory.  There is no
        external reference point. It's always considered cached.
        """
        super().__init__(**kwargs)

        # Store our raw config into memory
        self.content = content

        if self.config_format is None:
            # Detect our format if possible
            self.config_format = ConfigMemory.detect_config_format(
                self.content
            )

        return

    def url(self, privacy=False, *args, **kwargs):
        """Returns the URL built dynamically based on specified arguments."""

        return "memory://"

    def read(self, **kwargs):
        """Simply return content stored into memory."""

        return self.content

    @staticmethod
    def parse_url(url):
        """Memory objects have no parseable URL."""
        # These URLs can not be parsed
        return None



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import asyncio
from collections.abc import Generator
from datetime import tzinfo
from functools import partial
import re
from typing import Any, ClassVar, Optional, TypedDict, Union
from zoneinfo import ZoneInfo

from ..apprise_attachment import AppriseAttachment
from ..common import (
    NOTIFY_FORMATS,
    OVERFLOW_MODES,
    NotifyFormat,
    NotifyImageSize,
    NotifyType,
    OverflowMode,
    PersistentStoreMode,
)
from ..locale import Translatable, gettext_lazy as _
from ..persistent_store import PersistentStore
from ..url import URLBase
from ..utils.format import smart_split
from ..utils.parse import parse_bool
from ..utils.time import zoneinfo


class RequirementsSpec(TypedDict, total=False):
    """Defines our plugin requirements."""

    packages_required: Optional[Union[str, list[str]]]
    packages_recommended: Optional[Union[str, list[str]]]
    details: Optional[Translatable]


class NotifyBase(URLBase):
    """This is the base class for all notification services."""

    # An internal flag used to test the state of the plugin. If set to
    # False, then the plugin is not used.  Plugins can disable themselves
    # due to enviroment issues (such as missing libraries, or platform
    # dependencies that are not present).  By default all plugins are
    # enabled.
    enabled = True

    # The category allows for parent inheritance of this object to alter
    # this when it's function/use is intended to behave differently. The
    # following category types exist:
    #
    #  native: Is a native plugin written/stored in `apprise/plugins/Notify*`
    #  custom: Is a custom plugin written/stored in a users plugin directory
    #          that they loaded at execution time.
    category = "native"

    # Some plugins may require additional packages above what is provided
    # already by Apprise.
    #
    # Use this section to relay this information to the users of the script to
    # help guide them with what they need to know if they plan on using your
    # plugin.   The below configuration should otherwise accomodate all normal
    # situations and will not requrie any updating:
    requirements: ClassVar[RequirementsSpec] = {
        # Use the description to provide a human interpretable description of
        # what is required to make the plugin work. This is only nessisary
        # if there are package dependencies.  Setting this to default will
        # cause a general response to be returned.  Only set this if you plan
        # on over-riding the default.  Always consider language support here.
        # So before providing a value do the following in your code base:
        #
        #  from apprise.AppriseLocale import gettext_lazy as _
        #
        # 'details': _('My detailed requirements')
        "details": None,
        # Define any required packages needed for the plugin to run.  This is
        # an array of strings that simply look like lines residing in a
        # `requirements.txt` file...
        #
        # As an example, an entry may look like:
        # 'packages_required': [
        #   'cryptography < 3.4`,
        # ]
        "packages_required": [],
        # Recommended packages identify packages that are not required to make
        # your plugin work, but would improve it's use or grant it access to
        # full functionality (that might otherwise be limited).
        # Similar to `packages_required`, you would identify each entry in
        # the array as you would in a `requirements.txt` file.
        #
        #   - Do not re-provide entries already in the `packages_required`
        "packages_recommended": [],
    }

    # The services URL
    service_url = None

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = None

    # Most Servers do not like more then 1 request per 5 seconds, so 5.5 gives
    # us a safe play range. Override the one defined already in the URLBase
    request_rate_per_sec = 5.5

    # Allows the user to specify the NotifyImageSize object
    image_size = None

    # The maximum allowable characters allowed in the body per message
    body_maxlen = 32768

    # Defines the maximum allowable characters in the title; set this to zero
    # if a title can't be used. Titles that are not used but are defined are
    # automatically placed into the body
    title_maxlen = 250

    # Set the maximum line count; if this is set to anything larger then zero
    # the message (prior to it being sent) will be truncated to this number
    # of lines. Setting this to zero disables this feature.
    body_max_line_count = 0

    # Persistent storage default settings
    persistent_storage = True

    # Timezone Default; by setting it to None, the timezone detected
    # on the server is used
    timezone = None

    # Default Notify Format
    notify_format = NotifyFormat.TEXT

    # Default Overflow Mode
    overflow_mode = OverflowMode.UPSTREAM

    # Our default is to no not use persistent storage beyond in-memory
    # reference
    storage_mode = PersistentStoreMode.MEMORY

    # Default Emoji Interpretation
    interpret_emojis = False

    # Support Attachments; this defaults to being disabled.
    # Since apprise allows you to send attachments without a body or title
    # defined, by letting Apprise know the plugin won't support attachments
    # up front, it can quickly pass over and ignore calls to these end points.

    # You must set this to true if your application can handle attachments.
    # You must also consider a flow change to your notification if this is set
    # to True as well as now there will be cases where both the body and title
    # may not be set.  There will never be a case where a body, or attachment
    # isn't set in the same call to your notify() function.
    attachment_support = False

    # Default Title HTML Tagging
    # When a title is specified for a notification service that doesn't accept
    # titles, by default apprise tries to give a plesant view and convert the
    # title so that it can be placed into the body. The default is to just
    # use a <b> tag.  The below causes the <b>title</b> to get generated:
    default_html_tag_id = "b"

    # Here is where we define all of the arguments we accept on the url
    # such as: schema://whatever/?overflow=upstream&format=text
    # These act the same way as tokens except they are optional and/or
    # have default values set if mandatory. This rule must be followed
    template_args = dict(
        URLBase.template_args,
        **{
            "overflow": {
                "name": _("Overflow Mode"),
                "type": "choice:string",
                "values": OVERFLOW_MODES,
                # Provide a default
                "default": overflow_mode,
                # look up default using the following parent class value at
                # runtime. The variable name identified here (in this case
                # overflow_mode) is checked and it's result is placed over-top
                # of the 'default'. This is done because once a parent class
                # inherits this one, the overflow_mode already set as a default
                # 'could' be potentially over-ridden and changed to a different
                # value.
                "_lookup_default": "overflow_mode",
            },
            "format": {
                "name": _("Notify Format"),
                "type": "choice:string",
                "values": NOTIFY_FORMATS,
                # Provide a default
                "default": notify_format,
                # look up default using the following parent class value at
                # runtime.
                "_lookup_default": "notify_format",
            },
            "emojis": {
                "name": _("Interpret Emojis"),
                # SSL Certificate Authority Verification
                "type": "bool",
                # Provide a default
                "default": interpret_emojis,
                # look up default using the following parent class value at
                # runtime.
                "_lookup_default": "interpret_emojis",
            },
            "store": {
                "name": _("Persistent Storage"),
                # Use Persistent Storage
                "type": "bool",
                # Provide a default
                "default": persistent_storage,
                # look up default using the following parent class value at
                # runtime.
                "_lookup_default": "persistent_storage",
            },
            "tz": {
                "name": _("Timezone"),
                "type": "string",
                # Provide a default
                "default": timezone,
                # look up default using the following parent class value at
                # runtime.
                "_lookup_default": "timezone",
            },
        },
    )

    #
    # Overflow Defaults / Configuration applicable to SPLIT mode only
    #

    # Display Count  [X/X]
    #               ^^^^^^
    #               \\\\\\
    #               6 characters (space + count)
    # Display Count  [XX/XX]
    #               ^^^^^^^^
    #               \\\\\\\\
    #               8 characters (space + count)
    # Display Count  [XXX/XXX]
    #               ^^^^^^^^^^
    #               \\\\\\\\\\
    #               10 characters (space + count)
    # Display Count  [XXXX/XXXX]
    #               ^^^^^^^^^^^^
    #               \\\\\\\\\\\\
    #               12 characters (space + count)
    #
    # Given the above + some buffer we come up with the following:
    # If this value is exceeded, display counts automatically shut off
    overflow_max_display_count_width = 12

    # The number of characters to reserver for whitespace buffering
    # This is detected automatically, but you can enforce a value if
    # you desire:
    overflow_buffer = 0

    # the min accepted length of a title to allow for a counter display
    overflow_display_count_threshold = 130

    # Whether or not when over-flow occurs, if the title should be repeated
    # each time the message is split up
    #   - None: Detect
    #   - True: Always display title once
    #   - False: Display the title for each occurance
    overflow_display_title_once = None

    # If this is set to to True:
    #   The title_maxlen should be considered as a subset of the body_maxlen
    #    Hence: len(title) + len(body) should never be greater then body_maxlen
    #
    # If set to False, then there is no corrorlation between title_maxlen
    #  restrictions and that of body_maxlen
    overflow_amalgamate_title = False

    # Identifies the timezone to use;  if this is not over-ridden, then the
    # timezone defined in the AppriseAsset() object is used instead.  The
    # Below is expected to be in a ZoneInfo type already.  You can have this
    # automatically initialized by specifying ?tz= on the Apprise URLs
    __tzinfo = None

    def __init__(self, **kwargs):
        """Initialize some general configuration that will keep things
        consistent when working with the notifiers that will inherit this
        class."""

        super().__init__(**kwargs)

        # Store our interpret_emoji's setting
        # If asset emoji value is set to a default of True and the user
        #   specifies it to be false, this is accepted and False over-rides.
        #
        # If asset emoji value is set to a default of None, a user may
        #   optionally over-ride this and set it to True from the Apprise
        #   URL. ?emojis=yes
        #
        # If asset emoji value is set to a default of False, then all emoji's
        # are turned off (no user over-rides allowed)
        #

        # Our Persistent Storage object is initialized on demand
        self.__store = None

        # Take a default
        self.interpret_emojis = self.asset.interpret_emojis
        if "emojis" in kwargs:
            # possibly over-ride default
            self.interpret_emojis = bool(
                self.interpret_emojis in (None, True)
                and parse_bool(
                    kwargs.get("emojis", False),
                    default=NotifyBase.template_args["emojis"]["default"],
                )
            )

        if "format" in kwargs:
            value = kwargs["format"]
            try:
                self.notify_format = (
                    value if isinstance(value, NotifyFormat)
                    else NotifyFormat(value.lower())
                )

            except (AttributeError, ValueError):
                err = (
                    f"An invalid notification format ({value}) was "
                    "specified.")
                self.logger.warning(err)
                raise TypeError(err) from None

        if "tz" in kwargs:
            value = kwargs["tz"]
            self.__tzinfo = zoneinfo(value)
            if not self.__tzinfo:
                err = (
                    f"An invalid notification timezone ({value}) was "
                    "specified.")
                self.logger.warning(err)
                raise TypeError(err) from None

        if "overflow" in kwargs:
            value = kwargs["overflow"]
            try:
                self.overflow_mode = (
                    value if isinstance(value, OverflowMode)
                    else OverflowMode(value.lower())
                )

            except (AttributeError, ValueError):
                err = f"An invalid overflow method ({value}) was specified."
                self.logger.warning(err)
                raise TypeError(err) from None

        # Prepare our Persistent Storage switch
        self.persistent_storage = parse_bool(
            kwargs.get("store", NotifyBase.persistent_storage)
        )
        if not self.persistent_storage:
            # Enforce the disabling of cache (ortherwise defaults are use)
            self.url_identifier = False
            self.__cached_url_identifier = None

    def image_url(
        self,
        notify_type: NotifyType,
        image_size: Optional[NotifyImageSize] = None,
        logo: bool = False,
        extension: Optional[str] = None,
    ) -> Optional[str]:
        """Returns Image URL if possible."""

        image_size = self.image_size if image_size is None else image_size
        if not image_size:
            return None

        return self.asset.image_url(
            notify_type=notify_type,
            image_size=image_size,
            logo=logo,
            extension=extension,
        )

    def image_path(
        self,
        notify_type: NotifyType,
        extension: Optional[str] = None,
    ) -> Optional[str]:
        """Returns the path of the image if it can."""
        if not self.image_size:
            return None

        return self.asset.image_path(
            notify_type=notify_type,
            image_size=self.image_size,
            extension=extension,
        )

    def image_raw(
        self,
        notify_type: NotifyType,
        extension: Optional[str] = None,
    ) -> Optional[bytes]:
        """Returns the raw image if it can."""
        if not self.image_size:
            return None

        return self.asset.image_raw(
            notify_type=notify_type,
            image_size=self.image_size,
            extension=extension,
        )

    def color(
        self,
        notify_type: NotifyType,
        color_type: Optional[type] = None,
    ) -> Union[str, int, tuple[int, int, int]]:
        """Returns the html color (hex code) associated with the
        notify_type."""

        return self.asset.color(
            notify_type=notify_type,
            color_type=color_type,
        )

    def ascii(
        self,
        notify_type: NotifyType,
    ) -> str:
        """Returns the ascii characters associated with the notify_type."""

        return self.asset.ascii(
            notify_type=notify_type,
        )

    def notify(self, *args: Any, **kwargs: Any) -> bool:
        """Performs notification."""
        try:
            # Build a list of dictionaries that can be used to call send().
            send_calls = list(self._build_send_calls(*args, **kwargs))

        except TypeError:
            # Internal error
            return False

        else:
            # Loop through each call, one at a time. (Use a list rather than a
            # generator to call all the partials, even in case of a failure.)
            the_calls = [self.send(**kwargs2) for kwargs2 in send_calls]
            return all(the_calls)

    async def async_notify(self, *args: Any, **kwargs: Any) -> bool:
        """Performs notification for asynchronous callers."""
        try:
            # Build a list of dictionaries that can be used to call send().
            send_calls = list(self._build_send_calls(*args, **kwargs))

        except TypeError:
            # Internal error
            return False

        else:
            loop = asyncio.get_event_loop()

            # Wrap each call in a coroutine that uses the default executor.
            # TODO: In the future, allow plugins to supply a native
            # async_send() method.
            async def do_send(**kwargs2):
                send = partial(self.send, **kwargs2)
                result = await loop.run_in_executor(None, send)
                return result

            # gather() all calls in parallel.
            the_cors = (do_send(**kwargs2) for kwargs2 in send_calls)
            return all(await asyncio.gather(*the_cors))

    def _build_send_calls(
        self,
        body: Optional[str] = None,
        title: Optional[str] = None,
        notify_type: NotifyType = NotifyType.INFO,
        overflow: Optional[Union[str, OverflowMode]] = None,
        attach: Optional[Union[list[str], AppriseAttachment]] = None,
        body_format: Optional[NotifyFormat] = None,
        **kwargs: Any,
    ) -> Generator[dict[str, Any], None, None]:
        """Get a list of dictionaries that can be used to call send() or (in
        the future) async_send()."""

        if not self.enabled:
            # Deny notifications issued to services that are disabled
            msg = f"{self.service_name} is currently disabled on this system."
            self.logger.warning(msg)
            raise TypeError(msg)

        # Prepare attachments if required
        if attach is not None and not isinstance(attach, AppriseAttachment):
            try:
                attach = AppriseAttachment(attach, asset=self.asset)

            except TypeError:
                # bad attachments
                raise

            # Handle situations where the body is None
            body = body if body else ""

        elif not (body or attach):
            # If there is not an attachment at the very least, a body must be
            # present
            msg = "No message body or attachment was specified."
            self.logger.warning(msg)
            raise TypeError(msg)

        if not body and not self.attachment_support:
            # If no body was specified, then we know that an attachment
            # was.  This is logic checked earlier in the code.
            #
            # Knowing this, if the plugin itself doesn't support sending
            # attachments, there is nothing further to do here, just move
            # along.
            msg = (
                f"{self.service_name} does not support attachments; "
                " service skipped"
            )
            self.logger.warning(msg)
            raise TypeError(msg)

        # Handle situations where the title is None
        title = title if title else ""

        # Truncate flag set with attachments ensures that only 1
        # attachment passes through. In the event there could be many
        # services specified, we only want to do this logic once.
        # The logic is only applicable if ther was more then 1 attachment
        # specified
        overflow = self.overflow_mode if overflow is None else overflow
        if attach and len(attach) > 1 and overflow == OverflowMode.TRUNCATE:
            # Save first attachment
            _attach = AppriseAttachment(attach[0], asset=self.asset)
        else:
            # reference same attachment
            _attach = attach

        # Apply our overflow (if defined)
        for chunk in self._apply_overflow(
            body=body, title=title, overflow=overflow, body_format=body_format
        ):

            # Send notification
            yield {
                "body": chunk["body"],
                "title": chunk["title"],
                "notify_type": notify_type,
                "attach": _attach,
                "body_format": body_format,
            }

    def _apply_overflow(
        self,
        body: Optional[str],
        title: Optional[str] = None,
        overflow: Optional[Union[str, OverflowMode]] = None,
        body_format: Optional[NotifyFormat] = None,
    ) -> list[dict[str, str]]:
        """
        Apply overflow behaviour (UPSTREAM, TRUNCATE, SPLIT) to title/body.

        Takes the message body and title as input.  This function then
        applies any defined overflow restrictions associated with the
        notification service and may alter the message if/as required.

        The function will always return a list object in the following
        structure:
            [
                {
                    title: 'the title goes here',
                    body: 'the message body goes here',
                },
                {
                    title: 'the title goes here',
                    body: 'the continued message body goes here',
                },
            ]
        """
        response: list[dict[str, str]] = []

        # Tidy
        title = "" if not title else title.strip()
        body = "" if not body else body.rstrip()

        # Default overflow mode
        if overflow is None:
            overflow = self.overflow_mode

        # Default effective body format
        if body_format is None:
            body_format = self.notify_format

        # If the service does not support a title, amalgamate into body
        if self.title_maxlen <= 0 and len(title) > 0:
            if self.notify_format == NotifyFormat.HTML:
                body = (
                    f"<{self.default_html_tag_id}>{title}"
                    f"</{self.default_html_tag_id}>"
                    f"<br />\r\n{body}"
                )

            elif (
                self.notify_format == NotifyFormat.MARKDOWN
                and body_format == NotifyFormat.TEXT
            ):
                # Content is appended to body as markdown
                title = title.lstrip("\r\n \t\v\f#-")
                if title:
                    body = f"# {title}\r\n{body}"

            else:
                body = f"{title}\r\n{body}"

            title = ""

        # Enforce line count
        if self.body_max_line_count > 0:
            lines = re.split(r"\r*\n", body)
            body = "\r\n".join(lines[0 : self.body_max_line_count])

        # UPSTREAM mode: do not touch content
        if overflow == OverflowMode.UPSTREAM:
            response.append({"body": body, "title": title})
            return response

        # a value of '2' allows for the \r\n that is applied when amalgamating
        overflow_buffer = (
            max(2, self.overflow_buffer)
            if (self.title_maxlen == 0 and len(title))
            else self.overflow_buffer
        )

        #
        # TRUNCATE and SPLIT require sizing logic
        #

        # Handle situations where body and title are amalgamated
        title_maxlen = (
            self.title_maxlen
            if not self.overflow_amalgamate_title
            else min(
                len(title) + self.overflow_max_display_count_width,
                self.title_maxlen,
                self.body_maxlen,
            )
        )

        if len(title) > title_maxlen:
            # Truncate our title
            title = title[:title_maxlen].rstrip()

        # Compute body_maxlen as per legacy logic
        if (
            self.overflow_amalgamate_title
            and (self.body_maxlen - overflow_buffer) >= title_maxlen
        ):
            # status quo
            body_maxlen = (
                self.body_maxlen
                if not title
                else (self.body_maxlen - title_maxlen)
            ) - overflow_buffer
        else:
            # If the body fits, we're done
            body_maxlen = (
                self.body_maxlen
                if not self.overflow_amalgamate_title
                else (self.body_maxlen - overflow_buffer)
            )

        # If the body fits, we are done
        if body_maxlen > 0 and len(body) <= body_maxlen:
            response.append({"body": body, "title": title})
            return response

        # TRUNCATE mode: hard truncation (no smart-splitting)
        if overflow == OverflowMode.TRUNCATE:
            response.append({
                "body": body[:body_maxlen].lstrip("\r\n\x0b\x0c").rstrip(),
                "title": title,
            })
            return response

        #
        # SPLIT mode
        #

        # Detect if we only display our title once or not (legacy logic)
        if self.overflow_display_title_once is None:
            # Detect if we only display our title once or not:
            overflow_display_title_once = bool(
                self.overflow_amalgamate_title
                and body_maxlen < self.overflow_display_count_threshold
            )
        else:
            overflow_display_title_once = self.overflow_display_title_once

        # SPLIT mode with repeated title (with/without counter)
        if not overflow_display_title_once and not (
            # edge case: amalgamated title but no body space
            self.overflow_amalgamate_title
            and body_maxlen <= 0
        ):
            # Decide whether to show a counter (legacy condition)
            show_counter = (
                title
                and len(body) > body_maxlen
                and (
                    (
                        self.overflow_amalgamate_title
                        and body_maxlen >=
                        self.overflow_display_count_threshold
                    )
                    or (
                        not self.overflow_amalgamate_title
                        and title_maxlen >
                        self.overflow_display_count_threshold
                    )
                )
                and (
                    title_maxlen >
                    (self.overflow_max_display_count_width + overflow_buffer)
                    and self.title_maxlen >=
                    self.overflow_display_count_threshold
                )
            )

            effective_body_maxlen = body_maxlen
            if show_counter:
                # introduce padding for the counter
                effective_body_maxlen -= overflow_buffer

            # Use smart splitting instead of naive slicing
            chunks = smart_split(
                body,
                effective_body_maxlen,
                body_format,
            )
            count = len(chunks)

            template = ""
            if show_counter:
                digits = len(str(count))
                overflow_display_count_width = 4 + (digits * 2)

                if (
                    overflow_display_count_width
                    <= self.overflow_max_display_count_width
                ):
                    # Truncate title further if needed to make room for counter
                    if (
                        len(title)
                        > title_maxlen - overflow_display_count_width
                    ):
                        title = title[
                            : title_maxlen - overflow_display_count_width
                        ]
                    template = f" [{{:0{digits}d}}/{{:0{digits}d}}]"
                else:
                    # Too many messages; fall back to repeated title without
                    # counter displayed
                    show_counter = False

            response = []
            for idx, chunk_body in enumerate(chunks, start=1):
                suffix = template.format(idx, count) if show_counter else ""
                response.append({
                    "body": chunk_body.lstrip("\r\n\x0b\x0c").rstrip(),
                    "title": f"{title}{suffix}",
                })

        else:
            #
            # SPLIT mode, display title once and move on
            # (this covers both overflow_display_title_once=True
            #  and the edge case body_maxlen <= 0 with amalgamated title)
            #
            response = []
            consumed = 0
            remainder = body

            if body_maxlen > 0 and body:
                # First chunk uses body_maxlen (which already accounts for
                # the title)
                first_chunks = smart_split(
                    body,
                    body_maxlen,
                    body_format,
                )
                first_body = first_chunks[0] if first_chunks else ""
                consumed = len(first_body)
                remainder = body[consumed:]

                response.append({
                    "body": first_body.lstrip("\r\n\x0b\x0c").rstrip(),
                    "title": title,
                })

            else:
                # body_maxlen <= 0 or no body; send title only, still honouring
                # body
                response.append({
                    "body": "",
                    "title": title,
                })
                # remainder stays as full body; will be split below

            # Remaining chunks: no title, use the full body_maxlen of the
            # service
            if remainder:
                more_chunks = smart_split(
                    remainder,
                    self.body_maxlen,
                    body_format,
                )
                for chunk_body in more_chunks:
                    response.append({
                        "body": chunk_body.lstrip("\r\n\x0b\x0c").rstrip(),
                        "title": "",
                    })

        return response

    def send(
        self,
        body: str,
        title: str = "",
        notify_type: NotifyType = NotifyType.INFO,
        **kwargs: Any,
    ) -> bool:
        """Should preform the actual notification itself."""
        raise NotImplementedError(
            "send() is not implimented by the child class."
        )

    def url_parameters(
        self,
        *args: Any,
        **kwargs: Any,
    ) -> dict[str, Any]:
        """Provides a default set of parameters to work with.

        This can greatly simplify URL construction in the acommpanied url()
        function in all defined plugin services.
        """

        params = {
            "format": self.notify_format.value,
            "overflow": self.overflow_mode.value,
        }

        # Timezone Information (if ZoneInfo)
        if self.__tzinfo and isinstance(self.__tzinfo, ZoneInfo):
            params["tz"] = self.__tzinfo.key

        # Persistent Storage Setting
        if self.persistent_storage != NotifyBase.persistent_storage:
            params["store"] = "yes" if self.persistent_storage else "no"

        params.update(super().url_parameters(*args, **kwargs))

        # return default parameters
        return params

    @staticmethod
    def parse_url(
        url: str,
        verify_host: bool = True,
        plus_to_space: bool = False,
    ) -> Optional[dict[str, Any]]:
        """Parses the URL and returns it broken apart into a dictionary.

        This is very specific and customized for Apprise.


        Args:
            url (str): The URL you want to fully parse.
            verify_host (:obj:`bool`, optional): a flag kept with the parsed
                 URL which some child classes will later use to verify SSL
                 keys (if SSL transactions take place).  Unless under very
                 specific circumstances, it is strongly recomended that
                 you leave this default value set to True.

        Returns:
            A dictionary is returned containing the URL fully parsed if
            successful, otherwise None is returned.
        """
        results = URLBase.parse_url(
            url, verify_host=verify_host, plus_to_space=plus_to_space
        )

        if not results:
            # We're done; we failed to parse our url
            return results

        # Allow overriding the default format
        if "format" in results["qsd"]:
            results["format"] = results["qsd"].get("format", "").lower()
            if results["format"] not in NOTIFY_FORMATS:
                URLBase.logger.warning(
                    "Unsupported format specified "
                    f"{results['qsd']['format']!r}"
                )
                del results["format"]

        # Allow overriding the default overflow
        if "overflow" in results["qsd"]:
            results["overflow"] = results["qsd"].get("overflow", "").lower()
            if results["overflow"] not in OVERFLOW_MODES:
                URLBase.logger.warning(
                    "Unsupported overflow mode specified "
                    f"{results['qsd']['overflow']!r}"
                )
                del results["overflow"]

        # Allow emoji's override
        if "emojis" in results["qsd"]:
            results["emojis"] = parse_bool(results["qsd"].get("emojis"))
            # Store our persistent storage boolean

        # Allow overriding the default timezone
        if "tz" in results["qsd"]:
            results["tz"] = results["qsd"].get("tz", "")

        if "store" in results["qsd"]:
            results["store"] = results["qsd"]["store"]

        return results

    @staticmethod
    def parse_native_url(url: str) -> Optional[dict[str, Any]]:
        """This is a base class that can be optionally over-ridden by child
        classes who can build their Apprise URL based on the one provided by
        the notification service they choose to use.

        The intent of this is to make Apprise a little more userfriendly to
        people who aren't familiar with constructing URLs and wish to use the
        ones that were just provied by their notification serivice that they're
        using.

        This function will return None if the passed in URL can't be matched as
        belonging to the notification service. Otherwise this function should
        return the same set of results that parse_url() does.
        """
        return None

    @property
    def store(self):
        """Returns a pointer to our persistent store for use.

        The best use cases are:
         self.store.get('key')
         self.store.set('key', 'value')
         self.store.delete('key1', 'key2', ...)

        You can also access the keys this way:
         self.store['key']

        And clear them:
         del self.store['key']
        """
        if self.__store is None:
            # Initialize our persistent store for use
            self.__store = PersistentStore(
                namespace=self.url_id(),
                path=self.asset.storage_path,
                mode=self.asset.storage_mode,
            )

        return self.__store

    @property
    def tzinfo(self) -> tzinfo:
        """Returns our tzinfo file associated with this plugin if set
        otherwise the default timezone is returned.
        """
        return self.__tzinfo if self.__tzinfo else self.asset.tzinfo



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from itertools import chain
from json import dumps
import re

import requests

from ..common import NotifyImageSize, NotifyType
from ..locale import gettext_lazy as _
from ..utils.parse import parse_bool, parse_list, validate_regex
from .base import NotifyBase
from .discord import USER_ROLE_DETECTION_RE

# Used to break path apart into list of channels
CHANNEL_LIST_DELIM = re.compile(r"[ \t\r\n,#\\/]+")

CHANNEL_REGEX = re.compile(r"^\s*(\#|\%35)?(?P<channel>[0-9]+)", re.I)

# For API Details see:
# https://notifiarr.wiki/Client/Installation

# Another good example:
# https://notifiarr.wiki/en/Website/ \
#              Integrations/Passthrough#payload-example-1


class NotifyNotifiarr(NotifyBase):
    """A wrapper for Notifiarr Notifications."""

    # The default descriptive name associated with the Notification
    service_name = "Notifiarr"

    # The services URL
    service_url = "https://notifiarr.com/"

    # The default secure protocol
    secure_protocol = "notifiarr"

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = "https://github.com/caronc/apprise/wiki/Notify_notifiarr"

    # The Notification URL
    notify_url = "https://notifiarr.com/api/v1/notification/apprise"

    # Notifiarr Throttling (knowing in advance reduces 429 responses)
    # define('NOTIFICATION_LIMIT_SECOND_USER', 5);
    # define('NOTIFICATION_LIMIT_SECOND_PATRON', 15);

    # Throttle requests ever so slightly
    request_rate_per_sec = 0.04

    # Allows the user to specify the NotifyImageSize object
    image_size = NotifyImageSize.XY_256

    # Define object templates
    templates = ("{schema}://{apikey}/{targets}",)

    # Define our apikeys; these are the minimum apikeys required required to
    # be passed into this function (as arguments). The syntax appends any
    # previously defined in the base package and builds onto them
    template_tokens = dict(
        NotifyBase.template_tokens,
        **{
            "apikey": {
                "name": _("Token"),
                "type": "string",
                "required": True,
                "private": True,
            },
            "target_channel": {
                "name": _("Target Channel"),
                "type": "string",
                "prefix": "#",
                "map_to": "targets",
            },
            "targets": {
                "name": _("Targets"),
                "type": "list:string",
                "required": True,
            },
        },
    )

    # Define our template arguments
    template_args = dict(
        NotifyBase.template_args,
        **{
            "key": {
                "alias_of": "apikey",
            },
            "apikey": {
                "alias_of": "apikey",
            },
            "event": {
                "name": _("Discord Event ID"),
                "type": "int",
            },
            "image": {
                "name": _("Include Image"),
                "type": "bool",
                "default": False,
                "map_to": "include_image",
            },
            "source": {
                "name": _("Source"),
                "type": "string",
            },
            "from": {"alias_of": "source"},
            "to": {
                "alias_of": "targets",
            },
        },
    )

    def __init__(
        self,
        apikey=None,
        include_image=None,
        event=None,
        targets=None,
        source=None,
        **kwargs,
    ):
        """Initialize Notifiarr Object.

        headers can be a dictionary of key/value pairs that you want to
        additionally include as part of the server headers to post with
        """
        super().__init__(**kwargs)

        self.apikey = apikey
        if not self.apikey:
            msg = f"An invalid Notifiarr APIKey ({apikey}) was specified."
            self.logger.warning(msg)
            raise TypeError(msg)

        # Place a thumbnail image inline with the message body
        self.include_image = (
            include_image
            if isinstance(include_image, bool)
            else self.template_args["image"]["default"]
        )

        # Prepare our source (if set)
        self.source = validate_regex(source)

        self.event = 0
        if event:
            try:
                self.event = int(event)

            except (ValueError, TypeError):
                msg = (
                    "An invalid Notifiarr Discord Event ID "
                    f"({event}) was specified."
                )
                self.logger.warning(msg)
                raise TypeError(msg) from None

        # Prepare our targets
        self.targets = {
            "channels": [],
            "invalid": [],
        }

        for target in parse_list(targets):
            result = CHANNEL_REGEX.match(target)
            if result:
                # Store role information
                self.targets["channels"].append(int(result.group("channel")))
                continue

            self.logger.warning(
                f"Dropped invalid channel ({target}) specified.",
            )
            self.targets["invalid"].append(target)

        return

    @property
    def url_identifier(self):
        """Returns all of the identifiers that make this URL unique from
        another simliar one.

        Targets or end points should never be identified here.
        """
        return (
            self.secure_protocol if self.secure else self.protocol,
            self.apikey,
        )

    def url(self, privacy=False, *args, **kwargs):
        """Returns the URL built dynamically based on specified arguments."""

        # Define any URL parameters
        params = {
            "image": "yes" if self.include_image else "no",
        }

        if self.source:
            params["source"] = self.source

        if self.event:
            params["event"] = self.event

        # Extend our parameters
        params.update(self.url_parameters(privacy=privacy, *args, **kwargs))

        return "{schema}://{apikey}/{targets}?{params}".format(
            schema=self.secure_protocol,
            apikey=self.pprint(self.apikey, privacy, safe=""),
            targets="/".join([
                NotifyNotifiarr.quote(x, safe="+#@")
                for x in chain(
                    # Channels
                    [f"#{x}" for x in self.targets["channels"]],
                    # Pass along the same invalid entries as were provided
                    self.targets["invalid"],
                )
            ]),
            params=NotifyNotifiarr.urlencode(params),
        )

    def send(self, body, title="", notify_type=NotifyType.INFO, **kwargs):
        """Perform Notifiarr Notification."""

        if not self.targets["channels"]:
            # There were no services to notify
            self.logger.warning("There were no Notifiarr channels to notify.")
            return False

        # No error to start with
        has_error = False

        # Acquire image_url
        image_url = self.image_url(notify_type)

        # Define our mentions
        mentions = {
            "pingUser": [],
            "pingRole": [],
            "content": [],
        }

        # parse for user id's <@123> and role IDs <@&456>
        results = USER_ROLE_DETECTION_RE.findall(body)
        if results:
            for is_role, no, value in results:
                if value:
                    # @everybody, @admin, etc - unsupported
                    mentions["content"].append(f"@{value}")

                elif is_role:
                    mentions["pingRole"].append(no)
                    mentions["content"].append(f"<@&{no}>")

                else:  # is_user
                    mentions["pingUser"].append(no)
                    mentions["content"].append(f"<@{no}>")

        for _idx, channel in enumerate(self.targets["channels"]):
            # prepare Notifiarr Object
            payload = {
                "source": self.source if self.source else self.app_id,
                "type": notify_type,
                "notification": {
                    "update": bool(self.event),
                    "name": self.app_id,
                    "event": str(self.event) if self.event else "",
                },
                "discord": {
                    "color": self.color(notify_type),
                    "ping": {
                        # Only 1 user is supported, so truncate the rest
                        "pingUser": (
                            0
                            if not mentions["pingUser"]
                            else mentions["pingUser"][0]
                        ),
                        # Only 1 role is supported, so truncate the rest
                        "pingRole": (
                            0
                            if not mentions["pingRole"]
                            else mentions["pingRole"][0]
                        ),
                    },
                    "text": {
                        "title": title,
                        "content": (
                            ""
                            if not mentions["content"]
                            else "👉 " + " ".join(mentions["content"])
                        ),
                        "description": body,
                        "footer": self.app_desc,
                    },
                    "ids": {
                        "channel": channel,
                    },
                },
            }

            if self.include_image and image_url:
                payload["discord"]["text"]["icon"] = image_url
                payload["discord"]["images"] = {
                    "thumbnail": image_url,
                }

            if not self._send(payload):
                has_error = True

        return not has_error

    def _send(self, payload):
        """Send notification."""
        self.logger.debug(
            "Notifiarr POST URL:"
            f" {self.notify_url} (cert_verify={self.verify_certificate!r})"
        )
        self.logger.debug(f"Notifiarr Payload: {payload!s}")

        # Prepare HTTP Headers
        headers = {
            "User-Agent": self.app_id,
            "Content-Type": "application/json",
            "Accept": "text/plain",
            "X-api-Key": self.apikey,
        }

        # Always call throttle before any remote server i/o is made
        self.throttle()

        try:
            r = requests.post(
                self.notify_url,
                data=dumps(payload),
                headers=headers,
                verify=self.verify_certificate,
                timeout=self.request_timeout,
            )
            if r.status_code < 200 or r.status_code >= 300:
                # We had a problem
                status_str = NotifyNotifiarr.http_response_code_lookup(
                    r.status_code
                )

                self.logger.warning(
                    "Failed to send Notifiarr %s notification: %serror=%s.",
                    status_str,
                    ", " if status_str else "",
                    str(r.status_code),
                )

                self.logger.debug(f"Response Details:\r\n{r.content}")

                # Return; we're done
                return False

            else:
                self.logger.info("Sent Notifiarr notification.")

        except requests.RequestException as e:
            self.logger.warning(
                "A Connection error occurred sending Notifiarr "
                f"Chat notification to {self.host}."
            )
            self.logger.debug(f"Socket Exception: {e!s}")

            # Return; we're done
            return False

        return True

    def __len__(self):
        """Returns the number of targets associated with this notification."""
        targets = len(self.targets["channels"]) + len(self.targets["invalid"])
        return targets if targets > 0 else 1

    @staticmethod
    def parse_url(url):
        """Parses the URL and returns enough arguments that can allow us to re-
        instantiate this object."""
        results = NotifyBase.parse_url(url, verify_host=False)
        if not results:
            # We're done early as we couldn't load the results
            return results

        # Get channels
        results["targets"] = NotifyNotifiarr.split_path(results["fullpath"])

        if "event" in results["qsd"] and len(results["qsd"]["event"]):
            results["event"] = NotifyNotifiarr.unquote(results["qsd"]["event"])

        # Include images with our message
        results["include_image"] = parse_bool(
            results["qsd"].get("image", False)
        )

        # Track if we need to extract the hostname as a target
        host_is_potential_target = False

        if "source" in results["qsd"] and len(results["qsd"]["source"]):
            results["source"] = NotifyNotifiarr.unquote(
                results["qsd"]["source"]
            )

        elif "from" in results["qsd"] and len(results["qsd"]["from"]):
            results["source"] = NotifyNotifiarr.unquote(results["qsd"]["from"])

        # Set our apikey if found as an argument
        if "apikey" in results["qsd"] and len(results["qsd"]["apikey"]):
            results["apikey"] = NotifyNotifiarr.unquote(
                results["qsd"]["apikey"]
            )

            host_is_potential_target = True

        elif "key" in results["qsd"] and len(results["qsd"]["key"]):
            results["apikey"] = NotifyNotifiarr.unquote(results["qsd"]["key"])

            host_is_potential_target = True

        else:
            # Pop the first element (this is the api key)
            results["apikey"] = NotifyNotifiarr.unquote(results["host"])

        if host_is_potential_target is True and results["host"]:
            results["targets"].append(NotifyNotifiarr.unquote(results["host"]))

        # Support the 'to' variable so that we can support rooms this way too
        # The 'to' makes it easier to use yaml configuration
        if "to" in results["qsd"] and len(results["qsd"]["to"]):
            results["targets"] += list(
                filter(
                    bool,
                    CHANNEL_LIST_DELIM.split(
                        NotifyNotifiarr.unquote(results["qsd"]["to"])
                    ),
                )
            )

        return results



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from json import dumps
import re

import requests

from .. import exception
from ..common import NotifyType
from ..locale import gettext_lazy as _
from ..url import PrivacyMode
from ..utils.parse import is_phone_no, parse_bool, parse_phone_no
from .base import NotifyBase

GROUP_REGEX = re.compile(
    r"^\s*((\@|\%40)?(group\.)|\@|\%40)(?P<group>[a-z0-9_=-]+)", re.I
)


class NotifySignalAPI(NotifyBase):
    """A wrapper for SignalAPI Notifications."""

    # The default descriptive name associated with the Notification
    service_name = "Signal API"

    # The services URL
    service_url = "https://bbernhard.github.io/signal-cli-rest-api/"

    # The default protocol
    protocol = "signal"

    # The default protocol
    secure_protocol = "signals"

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = "https://github.com/caronc/apprise/wiki/Notify_signal"

    # Support attachments
    attachment_support = True

    # The maximum targets to include when doing batch transfers
    default_batch_size = 10

    # We don't support titles for Signal notifications
    title_maxlen = 0

    # Define object templates
    templates = (
        "{schema}://{host}/{from_phone}",
        "{schema}://{host}:{port}/{from_phone}",
        "{schema}://{user}@{host}/{from_phone}",
        "{schema}://{user}@{host}:{port}/{from_phone}",
        "{schema}://{user}:{password}@{host}/{from_phone}",
        "{schema}://{user}:{password}@{host}:{port}/{from_phone}",
        "{schema}://{host}/{from_phone}/{targets}",
        "{schema}://{host}:{port}/{from_phone}/{targets}",
        "{schema}://{user}@{host}/{from_phone}/{targets}",
        "{schema}://{user}@{host}:{port}/{from_phone}/{targets}",
        "{schema}://{user}:{password}@{host}/{from_phone}/{targets}",
        "{schema}://{user}:{password}@{host}:{port}/{from_phone}/{targets}",
    )

    # Define our template tokens
    template_tokens = dict(
        NotifyBase.template_tokens,
        **{
            "host": {
                "name": _("Hostname"),
                "type": "string",
                "required": True,
            },
            "port": {
                "name": _("Port"),
                "type": "int",
                "min": 1,
                "max": 65535,
            },
            "user": {
                "name": _("Username"),
                "type": "string",
            },
            "password": {
                "name": _("Password"),
                "type": "string",
                "private": True,
            },
            "from_phone": {
                "name": _("From Phone No"),
                "type": "string",
                "required": True,
                "regex": (r"^\+?[0-9\s)(+-]+$", "i"),
                "map_to": "source",
            },
            "target_phone": {
                "name": _("Target Phone No"),
                "type": "string",
                "prefix": "+",
                "regex": (r"^[0-9\s)(+-]+$", "i"),
                "map_to": "targets",
            },
            "target_channel": {
                "name": _("Target Group ID"),
                "type": "string",
                "prefix": "@",
                "regex": (r"^[a-z0-9_=-]+$", "i"),
                "map_to": "targets",
            },
            "targets": {
                "name": _("Targets"),
                "type": "list:string",
            },
        },
    )

    # Define our template arguments
    template_args = dict(
        NotifyBase.template_args,
        **{
            "to": {
                "alias_of": "targets",
            },
            "from": {
                "alias_of": "from_phone",
            },
            "batch": {
                "name": _("Batch Mode"),
                "type": "bool",
                "default": False,
            },
            "status": {
                "name": _("Show Status"),
                "type": "bool",
                "default": False,
            },
        },
    )

    def __init__(
        self, source=None, targets=None, batch=False, status=False, **kwargs
    ):
        """Initialize SignalAPI Object."""
        super().__init__(**kwargs)

        # Prepare Batch Mode Flag
        self.batch = batch

        # Set Status type
        self.status = status

        # Parse our targets
        self.targets = []

        # Used for URL generation afterwards only
        self.invalid_targets = []

        # Manage our Source Phone
        result = is_phone_no(source)
        if not result:
            msg = (
                "An invalid Signal API Source Phone No "
                f"({source}) was provided."
            )
            self.logger.warning(msg)
            raise TypeError(msg)

        self.source = "+{}".format(result["full"])

        if targets:
            # Validate our targerts
            for target in parse_phone_no(targets):
                # Validate targets and drop bad ones:
                result = is_phone_no(target)
                if result:
                    # store valid phone number
                    self.targets.append("+{}".format(result["full"]))
                    continue

                result = GROUP_REGEX.match(target)
                if result:
                    # Just store group information
                    self.targets.append(
                        "group.{}".format(result.group("group"))
                    )
                    continue

                self.logger.warning(
                    f"Dropped invalid phone/group ({target}) specified.",
                )
                self.invalid_targets.append(target)
                continue

        else:
            # Send a message to ourselves
            self.targets.append(self.source)

        return

    def send(
        self,
        body,
        title="",
        notify_type=NotifyType.INFO,
        attach=None,
        **kwargs,
    ):
        """Perform Signal API Notification."""

        if len(self.targets) == 0:
            # There were no services to notify
            self.logger.warning("There were no Signal API targets to notify.")
            return False

        # error tracking (used for function return)
        has_error = False

        attachments = []
        if attach and self.attachment_support:
            for attachment in attach:
                # Perform some simple error checking
                if not attachment:
                    # We could not access the attachment
                    self.logger.error(
                        "Could not access Signal API attachment"
                        f" {attachment.url(privacy=True)}."
                    )
                    return False

                try:
                    attachments.append(attachment.base64())

                except exception.AppriseException:
                    # We could not access the attachment
                    self.logger.error(
                        "Could not access Signal API attachment"
                        f" {attachment.url(privacy=True)}."
                    )
                    return False

                self.logger.debug(
                    "Appending Signal API attachment"
                    f" {attachment.url(privacy=True)}"
                )

        # Prepare our headers
        headers = {
            "User-Agent": self.app_id,
            "Content-Type": "application/json",
        }

        # Format defined here:
        #   https://bbernhard.github.io/signal-cli-rest-api\
        #       /#/Messages/post_v2_send
        # Example:
        # {
        #   "base64_attachments": [
        #     "string"
        #   ],
        #   "message": "string",
        #   "number": "string",
        #   "recipients": [
        #     "string"
        #   ]
        # }
        # Prepare our payload
        payload = {
            "message": (
                "{}{}".format(
                    (
                        ""
                        if not self.status
                        else f"{self.asset.ascii(notify_type)} "
                    ),
                    body,
                ).rstrip()
            ),
            "number": self.source,
            "recipients": [],
        }

        if attachments:
            # Store our attachments
            payload["base64_attachments"] = attachments

        # Determine Authentication
        auth = None
        if self.user:
            auth = (self.user, self.password)

        # Set our schema
        schema = "https" if self.secure else "http"

        # Construct our URL
        notify_url = f"{schema}://{self.host}"
        if isinstance(self.port, int):
            notify_url += f":{self.port}"
        notify_url += "/v2/send"

        # Send in batches if identified to do so
        batch_size = 1 if not self.batch else self.default_batch_size

        for index in range(0, len(self.targets), batch_size):
            # Prepare our recipients
            payload["recipients"] = self.targets[index : index + batch_size]

            self.logger.debug(
                "Signal API POST URL:"
                f" {notify_url} (cert_verify={self.verify_certificate!r})"
            )
            self.logger.debug(f"Signal API Payload: {payload!s}")

            # Always call throttle before any remote server i/o is made
            self.throttle()
            try:
                r = requests.post(
                    notify_url,
                    auth=auth,
                    data=dumps(payload),
                    headers=headers,
                    verify=self.verify_certificate,
                    timeout=self.request_timeout,
                )
                if r.status_code not in (
                    requests.codes.ok,
                    requests.codes.created,
                ):
                    # We had a problem
                    status_str = NotifySignalAPI.http_response_code_lookup(
                        r.status_code
                    )

                    self.logger.warning(
                        "Failed to send {} Signal API notification{}: "
                        "{}{}error={}.".format(
                            len(self.targets[index : index + batch_size]),
                            (
                                f" to {self.targets[index]}"
                                if batch_size == 1
                                else "(s)"
                            ),
                            status_str,
                            ", " if status_str else "",
                            r.status_code,
                        )
                    )

                    self.logger.debug(f"Response Details:\r\n{r.content}")

                    # Mark our failure
                    has_error = True
                    continue

                else:
                    self.logger.info(
                        "Sent {} Signal API notification{}.".format(
                            len(self.targets[index : index + batch_size]),
                            (
                                f" to {self.targets[index]}"
                                if batch_size == 1
                                else "(s)"
                            ),
                        )
                    )

            except requests.RequestException as e:
                self.logger.warning(
                    "A Connection error occured sending"
                    f" {len(self.targets[index:index + batch_size])} Signal"
                    " API notification(s)."
                )
                self.logger.debug(f"Socket Exception: {e!s}")

                # Mark our failure
                has_error = True
                continue

        return not has_error

    @property
    def url_identifier(self):
        """Returns all of the identifiers that make this URL unique from
        another simliar one.

        Targets or end points should never be identified here.
        """
        return (
            self.secure_protocol if self.secure else self.protocol,
            self.user,
            self.password,
            self.host,
            self.port,
            self.source,
        )

    def url(self, privacy=False, *args, **kwargs):
        """Returns the URL built dynamically based on specified arguments."""

        # Define any URL parameters
        params = {
            "batch": "yes" if self.batch else "no",
            "status": "yes" if self.status else "no",
        }

        # Extend our parameters
        params.update(self.url_parameters(privacy=privacy, *args, **kwargs))

        # Determine Authentication
        auth = ""
        if self.user and self.password:
            auth = "{user}:{password}@".format(
                user=NotifySignalAPI.quote(self.user, safe=""),
                password=self.pprint(
                    self.password, privacy, mode=PrivacyMode.Secret, safe=""
                ),
            )
        elif self.user:
            auth = "{user}@".format(
                user=NotifySignalAPI.quote(self.user, safe=""),
            )

        default_port = 443 if self.secure else 80

        # So we can strip out our own phone (if present); create a copy of our
        # targets
        if len(self.targets) == 1 and self.source in self.targets:
            targets = []

        elif len(self.targets) == 0:
            # invalid phone-no were specified
            targets = self.invalid_targets

        else:
            # append @ to non-phone number entries as they are groups
            # Remove group. prefix as well
            targets = [f"@{x[6:]}" if x[0] != "+" else x for x in self.targets]

        return "{schema}://{auth}{hostname}{port}/{src}/{dst}?{params}".format(
            schema=self.secure_protocol if self.secure else self.protocol,
            auth=auth,
            # never encode hostname since we're expecting it to be a valid one
            hostname=self.host,
            port=(
                ""
                if self.port is None or self.port == default_port
                else f":{self.port}"
            ),
            src=self.source,
            dst="/".join(
                [NotifySignalAPI.quote(x, safe="@+") for x in targets]
            ),
            params=NotifySignalAPI.urlencode(params),
        )

    def __len__(self):
        """Returns the number of targets associated with this notification."""
        #
        # Factor batch into calculation
        #
        batch_size = 1 if not self.batch else self.default_batch_size
        targets = len(self.targets)
        if batch_size > 1:
            targets = int(targets / batch_size) + (
                1 if targets % batch_size else 0
            )

        return targets

    @staticmethod
    def parse_url(url):
        """Parses the URL and returns enough arguments that can allow us to re-
        instantiate this object."""

        results = NotifyBase.parse_url(url, verify_host=False)
        if not results:
            # We're done early as we couldn't load the results
            return results

        # Get our entries; split_path() looks after unquoting content for us
        # by default
        results["targets"] = NotifySignalAPI.split_path(results["fullpath"])

        # The hostname is our authentication key
        results["apikey"] = NotifySignalAPI.unquote(results["host"])

        if "from" in results["qsd"] and len(results["qsd"]["from"]):
            results["source"] = NotifySignalAPI.unquote(results["qsd"]["from"])

        elif results["targets"]:
            # The from phone no is the first entry in the list otherwise
            results["source"] = results["targets"].pop(0)

        # Support the 'to' variable so that we can support targets this way too
        # The 'to' makes it easier to use yaml configuration
        if "to" in results["qsd"] and len(results["qsd"]["to"]):
            results["targets"] += NotifySignalAPI.parse_phone_no(
                results["qsd"]["to"]
            )

        # Get Batch Mode Flag
        results["batch"] = parse_bool(results["qsd"].get("batch", False))

        # Get status switch
        results["status"] = parse_bool(results["qsd"].get("status", False))

        return results



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

__title__ = "Apprise"
__description__: str = \
    "Push Notifications that work with just about every platform!"
__version__ = "1.9.6"
__author__ = "Chris Caron"
__email__ = "lead2gold@gmail.com"
__license__ = "BSD 2-Clause"
__copyright__ = "Copyright (C) 2025 Chris Caron <lead2gold@gmail.com>"
__status__ = "Production"

from . import decorators, exception
from .apprise import Apprise
from .apprise_attachment import AppriseAttachment
from .apprise_config import AppriseConfig
from .asset import AppriseAsset
from .attachment.base import AttachBase
from .common import (
    CONFIG_FORMATS,
    CONTENT_INCLUDE_MODES,
    CONTENT_LOCATIONS,
    NOTIFY_FORMATS,
    NOTIFY_IMAGE_SIZES,
    NOTIFY_TYPES,
    OVERFLOW_MODES,
    PERSISTENT_STORE_MODES,
    PERSISTENT_STORE_STATES,
    ConfigFormat,
    ContentIncludeMode,
    ContentLocation,
    NotifyFormat,
    NotifyImageSize,
    NotifyType,
    OverflowMode,
    PersistentStoreMode,
)
from .config.base import ConfigBase
from .locale import AppriseLocale

# Inherit our logging with our additional entries added to it
from .logger import LogCapture, logger, logging
from .manager_attachment import AttachmentManager
from .manager_config import ConfigurationManager
from .manager_plugins import NotificationManager
from .persistent_store import PersistentStore
from .plugins.base import NotifyBase
from .url import PrivacyMode, URLBase

# Set default logging handler to avoid "No handler found" warnings.
logging.getLogger(__name__).addHandler(logging.NullHandler())

__all__ = [
    "CONFIG_FORMATS",
    "CONTENT_INCLUDE_MODES",
    "CONTENT_LOCATIONS",
    "NOTIFY_FORMATS",
    "NOTIFY_IMAGE_SIZES",
    "NOTIFY_TYPES",
    "OVERFLOW_MODES",
    "PERSISTENT_STORE_MODES",
    "PERSISTENT_STORE_STATES",
    # Core
    "Apprise",
    "AppriseAsset",
    "AppriseAttachment",
    "AppriseConfig",
    "AppriseLocale",
    "AttachBase",
    "AttachmentManager",
    "ConfigBase",
    "ConfigFormat",
    "ConfigurationManager",
    "ContentIncludeMode",
    "ContentLocation",
    "LogCapture",
    # Managers
    "NotificationManager",
    "NotifyBase",
    "NotifyFormat",
    "NotifyImageSize",
    # Reference
    "NotifyType",
    "OverflowMode",
    "PersistentStore",
    "PersistentStoreMode",
    "PrivacyMode",
    "URLBase",
    # Decorator
    "decorators",
    # Exceptions
    "exception",
    # Logging
    "logger",
    "logging",
]



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from itertools import chain
from json import dumps
import re

import requests

from ..common import NotifyType
from ..locale import gettext_lazy as _
from ..url import PrivacyMode
from ..utils.parse import parse_list, validate_regex
from .base import NotifyBase

# Used to detect and parse channels
IS_CHANNEL = re.compile(r"^#?(?P<name>[A-Za-z0-9]+)$")

# Used to detect and parse a users push id
IS_USER_PUSHED_ID = re.compile(r"^@(?P<name>[A-Za-z0-9]+)$")


class NotifyPushed(NotifyBase):
    """A wrapper to Pushed Notifications."""

    # The default descriptive name associated with the Notification
    service_name = "Pushed"

    # The services URL
    service_url = "https://pushed.co/"

    # The default secure protocol
    secure_protocol = "pushed"

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = "https://github.com/caronc/apprise/wiki/Notify_pushed"

    # Pushed uses the http protocol with JSON requests
    notify_url = "https://api.pushed.co/1/push"

    # A title can not be used for Pushed Messages.  Setting this to zero will
    # cause any title (if defined) to get placed into the message body.
    title_maxlen = 0

    # The maximum allowable characters allowed in the body per message
    body_maxlen = 160

    # Define object templates
    templates = (
        "{schema}://{app_key}/{app_secret}",
        "{schema}://{app_key}/{app_secret}@{targets}",
    )

    # Define our template tokens
    template_tokens = dict(
        NotifyBase.template_tokens,
        **{
            "app_key": {
                "name": _("Application Key"),
                "type": "string",
                "private": True,
                "required": True,
            },
            "app_secret": {
                "name": _("Application Secret"),
                "type": "string",
                "private": True,
                "required": True,
            },
            "target_user": {
                "name": _("Target User"),
                "prefix": "@",
                "type": "string",
                "map_to": "targets",
            },
            "target_channel": {
                "name": _("Target Channel"),
                "type": "string",
                "prefix": "#",
                "map_to": "targets",
            },
            "targets": {
                "name": _("Targets"),
                "type": "list:string",
            },
        },
    )

    # Define our template arguments
    template_args = dict(
        NotifyBase.template_args,
        **{
            "to": {
                "alias_of": "targets",
            },
        },
    )

    def __init__(self, app_key, app_secret, targets=None, **kwargs):
        """Initialize Pushed Object."""
        super().__init__(**kwargs)

        # Application Key (associated with project)
        self.app_key = validate_regex(app_key)
        if not self.app_key:
            msg = (
                f"An invalid Pushed Application Key ({app_key}) was specified."
            )
            self.logger.warning(msg)
            raise TypeError(msg)

        # Access Secret (associated with project)
        self.app_secret = validate_regex(app_secret)
        if not self.app_secret:
            msg = (
                "An invalid Pushed Application Secret "
                f"({app_secret}) was specified."
            )
            self.logger.warning(msg)
            raise TypeError(msg)

        # Initialize channel list
        self.channels = []

        # Initialize user list
        self.users = []

        # Get our targets
        targets = parse_list(targets)
        if targets:
            # Validate recipients and drop bad ones:
            for target in targets:
                result = IS_CHANNEL.match(target)
                if result:
                    # store valid device
                    self.channels.append(result.group("name"))
                    continue

                result = IS_USER_PUSHED_ID.match(target)
                if result:
                    # store valid room
                    self.users.append(result.group("name"))
                    continue

                self.logger.warning(
                    f"Dropped invalid channel/userid ({target}) specified.",
                )

            if len(self.channels) + len(self.users) == 0:
                # We have no valid channels or users to notify after
                # explicitly identifying at least one.
                msg = "No Pushed targets to notify."
                self.logger.warning(msg)
                raise TypeError(msg)

        return

    def send(self, body, title="", notify_type=NotifyType.INFO, **kwargs):
        """Perform Pushed Notification."""

        # Initiaize our error tracking
        has_error = False

        # prepare JSON Object
        payload = {
            "app_key": self.app_key,
            "app_secret": self.app_secret,
            "target_type": "app",
            "content": body,
        }

        # So the logic is as follows:
        #  - if no user/channel was specified, then we just simply notify the
        #    app.
        #  - if there are user/channels specified, then we only alert them
        #    while respecting throttle limits (in the event there are a lot of
        #    entries.

        if len(self.channels) + len(self.users) == 0:
            # Just notify the app
            return self._send(
                payload=payload, notify_type=notify_type, **kwargs
            )

        # If our code reaches here, we want to target channels and users (by
        # their Pushed_ID instead...

        # Generate a copy of our original list
        channels = list(self.channels)
        users = list(self.users)

        # Copy our payload
        _payload = dict(payload)
        _payload["target_type"] = "channel"

        while len(channels) > 0:
            # Get Channel
            _payload["target_alias"] = channels.pop(0)

            if not self._send(
                payload=_payload, notify_type=notify_type, **kwargs
            ):

                # toggle flag
                has_error = True

        # Copy our payload
        _payload = dict(payload)
        _payload["target_type"] = "pushed_id"

        # Send all our defined User Pushed ID's
        while len(users):
            # Get User's Pushed ID
            _payload["pushed_id"] = users.pop(0)

            if not self._send(
                payload=_payload, notify_type=notify_type, **kwargs
            ):

                # toggle flag
                has_error = True

        return not has_error

    def _send(self, payload, notify_type, **kwargs):
        """A lower level call that directly pushes a payload to the Pushed
        Notification servers.

        This should never be called directly; it is referenced automatically
        through the send() function.
        """

        headers = {
            "User-Agent": self.app_id,
            "Content-Type": "application/json",
        }

        self.logger.debug(
            "Pushed POST URL:"
            f" {self.notify_url} (cert_verify={self.verify_certificate!r})"
        )
        self.logger.debug(f"Pushed Payload: {payload!s}")

        # Always call throttle before any remote server i/o is made
        self.throttle()

        try:
            r = requests.post(
                self.notify_url,
                data=dumps(payload),
                headers=headers,
                verify=self.verify_certificate,
                timeout=self.request_timeout,
            )

            if r.status_code != requests.codes.ok:
                # We had a problem
                status_str = NotifyPushed.http_response_code_lookup(
                    r.status_code
                )

                self.logger.warning(
                    "Failed to send Pushed notification:{}{}error={}.".format(
                        status_str, ", " if status_str else "", r.status_code
                    )
                )

                self.logger.debug(f"Response Details:\r\n{r.content}")

                # Return; we're done
                return False

            else:
                self.logger.info("Sent Pushed notification.")

        except requests.RequestException as e:
            self.logger.warning(
                "A Connection error occurred sending Pushed notification."
            )
            self.logger.debug(f"Socket Exception: {e!s}")

            # Return; we're done
            return False

        return True

    @property
    def url_identifier(self):
        """Returns all of the identifiers that make this URL unique from
        another simliar one.

        Targets or end points should never be identified here.
        """
        return (self.secure_protocol, self.app_key, self.app_secret)

    def url(self, privacy=False, *args, **kwargs):
        """Returns the URL built dynamically based on specified arguments."""

        # Our URL parameters
        params = self.url_parameters(privacy=privacy, *args, **kwargs)

        return "{schema}://{app_key}/{app_secret}/{targets}/?{params}".format(
            schema=self.secure_protocol,
            app_key=self.pprint(self.app_key, privacy, safe=""),
            app_secret=self.pprint(
                self.app_secret, privacy, mode=PrivacyMode.Secret, safe=""
            ),
            targets="/".join([
                NotifyPushed.quote(x)
                for x in chain(
                    # Channels are prefixed with a pound/hashtag symbol
                    [f"#{x}" for x in self.channels],
                    # Users are prefixed with an @ symbol
                    [f"@{x}" for x in self.users],
                )
            ]),
            params=NotifyPushed.urlencode(params),
        )

    def __len__(self):
        """Returns the number of targets associated with this notification."""
        targets = len(self.channels) + len(self.users)
        return targets if targets > 0 else 1

    @staticmethod
    def parse_url(url):
        """Parses the URL and returns enough arguments that can allow us to re-
        instantiate this object."""
        results = NotifyBase.parse_url(url, verify_host=False)
        if not results:
            # We're done early as we couldn't load the results
            return results

        # The first token is stored in the hostname
        app_key = NotifyPushed.unquote(results["host"])

        entries = NotifyPushed.split_path(results["fullpath"])
        # Now fetch the remaining tokens
        try:
            app_secret = entries.pop(0)

        except IndexError:
            # Force some bad values that will get caught
            # in parsing later
            app_secret = None
            app_key = None

        # Get our recipients (based on remaining entries)
        results["targets"] = entries

        # The 'to' makes it easier to use yaml configuration
        if "to" in results["qsd"] and len(results["qsd"]["to"]):
            results["targets"] += NotifyPushed.parse_list(results["qsd"]["to"])

        results["app_key"] = app_key
        results["app_secret"] = app_secret

        return results
